# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""
Compute ion concentrations with general Poisson-Nernst-Planck (PNP) equations.

Copyright 2019 IMTEK Simulation
University of Freiburg

Authors:
  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
"""
import logging
import time
import numpy as np
import scipy.constants as sc
import scipy.optimize

logger = logging.getLogger(__name__)

# Druecke nicht-linearen Teil der Transportgleichung (genauer, des Flusses) ueber
# Bernoulli-Funktionen
#
# $$ B(x) = \frac{x}{\exp(x)-1} $$
#
# aus. Damit wir in der Naehe von 0 nicht "in die Bredouille geraten", verwenden
# wir hier lieber die Taylorentwicklung. In der Literatur (Selbherr, S. Analysis
# and Simulation of Semiconductor Devices, Spriger 1984) wird eine noch
# aufwendigere stueckweise Definition empfohlen, allerdings werden wir im
# Folgenden sehen, dass unser Ansatz fuer dieses stationaere Problem genuegt.


def B(x):
    """Bernoulli function."""
    return np.where(
        np.abs(x) < 1e-9,
        1 - x/2 + x**2/12 - x**4/720,  # Taylor
        x / (np.exp(x) - 1))


# "lazy" Ansatz for approximating Jacobian
def jacobian(f, x0, dx=np.NaN):
    """Naive way to construct N x N Jacobin Fij from N-valued function
    f of N-valued vector x0.

    Parameters
    ----------
    f : callable
        N-valued function of N-valued variable vector
    x0 : (N,) ndarray
        N-valued variable vector
    dx : float (default: np.nan)
        Jacobian built with finite difference scheme of spacing ``dx``.
        If ``np.nan``, then use machine precision.

    Returns
    -------
    F : (N,N) ndarray
        NxN-valued 2nd order finite difference scheme approximate of Jacobian
        convention: F_ij = dfidxj, where i are array rows, j are array columns
    """


    N = len(x0)
    # choose step as small as possible
    if np.isnan(dx).any():
        res = np.finfo('float64').resolution
        dx = np.abs(x0) * np.sqrt(res)
        dx[dx < res] = res

    if np.isscalar(dx):
        dx = np.ones(N) * dx

    F = np.zeros((N,N)) # Jacobian Fij

    # convention: dfi_dxj
    # i are rows, j are columns
    for j in range(N):
        dxj = np.zeros(N)
        dxj[j] = dx[j]

        F[:,j] =  (f(x0 + dxj) - f(x0 - dxj)) / (2.0*dxj[j])

    return F

class PoissonNernstPlanckSystem:
    """Describes and solves a 1D Poisson-Nernst-Planck system"""

    # properties "offer" the solution in physical units:
    @property
    def grid(self):
        return self.X*self.l_unit

    @property
    def potential(self):
        return self.uij*self.u_unit

    @property
    def concentration(self):
        return np.where(self.nij > np.finfo('float64').resolution,
            self.nij*self.c_unit, 0.0)

    @property
    def charge_density(self):
        return np.sum(self.F * self.concentration.T * self.z,axis=1)

    @property
    def x1_scaled(self):
        return self.x0_scaled + self.L_scaled

    #TODO:  replace "didactic" Newton solver from IMTEK Simulation course with
    #       some standard package
    def newton(self,f,xij,**kwargs):
        """Newton solver expects system f and initial value xij

        Parameters
        ----------
        f : callable
            N-valued function of N-valued vector
        xij : (N,) ndarray
            N-valued initial value vector

        Returns
        -------
        xij : (N,) ndarray
            N-valued solution vector
        """
        self.xij = []
        self.converged = True
        # assume convergence, set to 'false' if maxit exceeded later

        self.logger.debug('Newton solver, grid points N = {:d}'.format(self.N))
        self.logger.debug('Newton solver, tolerance e = {:> 8.4g}'.format(self.e))
        self.logger.debug('Newton solver, maximum number of iterations M = {:d}'.format(self.maxit))

        i = 0
        delta_rel = 2*self.e

        self.logger.info("Convergence criterion: norm(dx) < {:4.2e}".format(self.e))

        self.convergenceStepAbsolute = np.zeros(self.maxit)
        self.convergenceStepRelative = np.zeros(self.maxit)
        self.convergenceResidualAbsolute = np.zeros(self.maxit)

        dxij = np.zeros(self.N)
        while delta_rel > self.e and i < self.maxit:
            self.logger.debug('*** Newton solver iteration {:d} ***'.format(i))

            # avoid cluttering log
            self.logger.disabled = True
            J = jacobian(f, xij)
            self.logger.disabled = False

            rank = np.linalg.matrix_rank(J)
            self.logger.debug('    Jacobian ({}) rank {:d}'.format(J.shape, rank))

            if rank < self.N:
                self.logger.warn("Singular jacobian of rank"
                      + "{:d} < {:d} at step {:d}".format(
                      rank, self.N, i ))
                break

            F = f(xij)
            invJ = np.linalg.inv(J)

            dxij = np.dot( invJ, F )

            delta = np.linalg.norm(dxij)
            delta_rel = delta / np.linalg.norm(xij)

            xij -= dxij
            self.xij.append(xij)

            normF = np.linalg.norm(F)

            self.logger.debug('    convergence norm(dxij), absolute {:> 8.4g}'.format(delta))
            self.logger.debug('    convergence norm(dxij), realtive {:> 8.4g}'.format(delta_rel))
            self.logger.debug('          residual norm(F), absolute {:> 8.4g}'.format(normF))

            self.convergenceStepAbsolute[i] = delta
            self.convergenceStepRelative[i] = delta_rel
            self.convergenceResidualAbsolute[i] = normF
            self.logger.info("Step {:4d}: norm(dx)/norm(x) = {:4.2e}, norm(dx) = {:4.2e}, norm(F) = {:4.2e}".format(
                i, delta_rel, delta, normF) )

            i += 1

        if i == self.maxit:
            self.logger.warn("Maximum number of iterations reached")
            self.converged = False

        self.logger.info("Ended after {:d} steps.".format(i))

        self.convergenceStepAbsolute = self.convergenceStepAbsolute[:i]
        self.convergenceStepRelative = self.convergenceStepRelative[:i]
        self.convergenceResidualAbsolute = self.convergenceResidualAbsolute[:i]
        return xij

    def solver_callback(self, xij, *_):
        """Callback function that can be used by optimizers of scipy.optimize.
        The second argument "*_" makes sure that it still works when the
        optimizer calls the callback function with more than one argument. See
        https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
        """
        if self.callback_count == 0:
            logger.info(
                "{:>12s} {:>12s} {:>12s} {:>12s} {:>12s} {:>12s}".format(
                    "#callback","residual norm","abs dx norm", "rel dx norm",
                    "timing, step", "timing, tot.") )
            self.xij = [ self.xi0 ]

            self.converged = True # TODO remove (?)
            self.convergenceStepAbsolute = []
            self.convergenceStepRelative = []
            self.convergenceResidualAbsolute = []


            dxij = np.zeros(self.N)

        self.xij.append(xij)
        dxij = xij - self.xij[self.callback_count]

        delta = np.linalg.norm(dxij)
        delta_rel = delta / np.linalg.norm(xij)

        fj = self.G(xij)
        norm_fj = np.linalg.norm(fj)

        self.convergenceStepAbsolute.append(delta)
        self.convergenceStepRelative.append(delta_rel)
        self.convergenceResidualAbsolute.append(norm_fj)

        t1 = time.perf_counter()
        dt = t1 - self.tj
        dT = t1 - self.t0
        self.tj = t1

        logger.info(
            "{:12d} {:12.5e} {:12.5e} {:12.5e} {:12.5e} {:12.5e}".format(
                self.callback_count, norm_fj, delta , delta_rel, dt, dT))

        self.callback_count += 1
        return


    def discretize(self):
        """Sets up discretization scheme and initial value"""
        # indices
        self.Ni = self.N+1
        I = np.arange(self.Ni)

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'discretization segments N', self.N, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'grid points N', self.Ni, lwidth=self.label_width))

        # discretization
        self.dx      = self.L_scaled / self.N # spatial step

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
          'dx', self.dx, lwidth=self.label_width))

        # positions (scaled)
        self.X = self.x0_scaled + I*self.dx

        # Bounary & initial values

        # internally:
        #   n: dimensionless concentrations
        #   u: dimensionless potential
        #   i: spatial index
        #   j: temporal index
        #   k: species index
        # initial concentrations equal to bulk concentrations

        # Kronecker product, M rows (ion species), Ni cols (grid points),
        if self.ni0 is None:
            self.ni0 = np.kron( self.c_scaled, np.ones((self.Ni,1)) ).T
        if self.zi0 is None:
            self.zi0 = np.kron( self.z, np.ones((self.Ni,1)) ).T # does not change

        # self.initial_values()

    def initial_values(self):
        """
        Solves decoupled linear system to get inital potential distribution.
        """

        zini0 = self.zi0*self.ni0 # z*n
        # shape: ion species (rows), grid points (cols), sum over ion species (along rows)
        rhoi0 = 0.5*zini0.sum(axis=0)

        # system matrix of spatial poisson equation
        Au = np.zeros((self.Ni,self.Ni))
        bu = np.zeros(self.Ni)
        Au[0,0]   = 1
        Au[-1,-1] = 1
        for i in range(1,self.N):
            Au[i,[i-1,i,i+1]] = [1.0, -2.0, 1.0] # 1D Laplace operator, 2nd order

        bu = rhoi0*self.dx**2 # => Poisson equation
        bu[0]  = self.u0
        bu[-1] = self.u1

        # get initial potential distribution by solving Poisson equation
        self.ui0 = np.dot( np.linalg.inv(Au), bu) # A u - b = 0 <=> u = A^-1 b

        return self.ui0

    # evokes Newton solver
    def solve(self):
        """Evokes solver

        Returns
        -------
        uij : (Ni,) ndarray
            potential at Ni grid points
        nij : (M,Nij) ndarray
            concentrations of M species at Ni grid points
        lamj: (L,) ndarray
            value of L Lagrange multipliers (not implemented, empty)
        """

        # if not yet done, set up initial values
        if self.ui0 is None:
            self.initial_values()

        if len(self.g) > 0:
            self.xi0 = np.concatenate([self.ui0, self.ni0.flatten(), np.zeros(len(self.g))])
        else:
            self.xi0 = np.concatenate([self.ui0, self.ni0.flatten()])

        self.callback_count = 0
        self.t0 = time.perf_counter()
        self.tj = self.t0 # previosu callback timer value

        # neat lecture on scipy optimizers
        # http://scipy-lectures.org/advanced/mathematical_optimization/
        if isinstance(self.solver, str) and self.solver in [
            'hybr','lm','broyden1','broyden2','anderson','linearmixing',
            'diagbroyden','excitingmixing','krylov','df-sane']:
            res = scipy.optimize.root(self.G,self.xi0,
                method=self.solver,callback=self.solver_callback,
                options = self.options)
            self.xij1 = res.x
            if not res.success:
                logger.warn(res.message)
        elif isinstance( self.solver, str):
            f = lambda x: np.linalg.norm(self.G(x))
            res = scipy.optimize.minimize(f,self.xi0.copy(),
                method=self.solver,callback=self.solver_callback,
                options=self.options)
            self.xij1 = res.x
            if not res.success:
                logger.warn(res.message)
        else:
            self.xij1 = self.solver(self.G,self.xi0.copy(),
                callback=self.solver_callback, options=self.options)

        # store results:
        self.uij  = self.xij1[:self.Ni] # potential
        self.nij  = self.xij1[self.Ni:(self.M+1)*self.Ni].reshape(self.M, self.Ni) # concentrations
        self.lamj = self.xij1[(self.M+1)*self.Ni:] # Lagrange multipliers

        return self.uij, self.nij, self.lamj

    # standard sets of boundary conditions:
    def useStandardInterfaceBC(self):
        """Interface at left hand side and open bulk at right hand side"""
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled
        self.u1 = 0
        self.logger.info('Left hand side Dirichlet boundary condition:                               u0 = {:> 8.4g}'.format(self.u0))
        self.logger.info('Right hand side Dirichlet boundary condition:                              u1 = {:> 8.4g}'.format(self.u1))

        self.boundary_conditions.extend([
            lambda x: self.leftPotentialDirichletBC(x,self.u0),
            lambda x: self.rightPotentialDirichletBC(x,self.u1) ])
        # self.rightPotentialBC = lambda x: self.rightPotentialDirichletBC(x,self.u1)

        #self.rightConcentrationBC = []
        for k in range(self.M):
          self.logger.info('Ion species {:02d} left hand side concentration Flux boundary condition:       j0 = {:> 8.4g}'.format(k,0))
          self.logger.info('Ion species {:02d} right hand side concentration Dirichlet boundary condition: c1 = {:> 8.4g}'.format(k,self.c_scaled[k]))
          self.boundary_conditions.extend( [
            lambda x, k=k: self.leftControlledVolumeSchemeFluxBC(x,k),
            lambda x, k=k: self.rightDirichletBC(x,k,self.c_scaled[k]) ] )
          #self.rightConcentrationBC.append(
          #  lambda x, k=k: self.rightDirichletBC(x,k,self.c_scaled[k]) )
        # counter-intuitive behavior of lambda in loop:
        # https://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture
        # workaround: default parameter k=k

    def useStandardCellBC(self):
        """Interfaces at left hand side and right hand side"""
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.0
        self.u1 = - self.delta_u_scaled / 2.
        self.logger.info('{:>{lwidth}s} u0 = {:< 8.4g}'.format(
          'Left hand side Dirichlet boundary condition', self.u0, lwidth=self.label_width))
        self.logger.info('{:>{lwidth}s} u1 = {:< 8.4g}'.format(
          'Right hand side Dirichlet boundary condition', self.u1, lwidth=self.label_width))
        self.boundary_conditions.extend([
          lambda x: self.leftPotentialDirichletBC(x,self.u0),
          lambda x: self.rightPotentialDirichletBC(x,self.u1) ])

        N0 = self.L_scaled*self.c_scaled # total amount of species in cell
        for k in range(self.M):
          self.logger.info('{:>{lwidth}s} j0 = {:<8.4g}'.format(
            'Ion species {:02d} left hand side concentration Flux boundary condition'.format(k),
            0.0, lwidth=self.label_width))
          self.logger.info('{:>{lwidth}s} N0 = {:<8.4g}'.format(
            'Ion species {:02d} number conservation constraint'.format(k),
            N0[k], lwidth=self.label_width))

          self.boundary_conditions.extend(  [
            lambda x, k=k: self.leftControlledVolumeSchemeFluxBC(x,k),
            lambda x, k=k, N0=N0[k]: self.numberConservationConstraint(x,k,N0) ] )

    def useSternLayerCellBC(self, implicit=False):
        """Interfaces at left hand side and right hand side,
        Stern layer either by prescribing linear potential regime between cell
        boundary and outer Helmholtz plane (OHP), or by applying Robin BC;
        zero flux BC on all ion species.

        Parameters
        ----------
        implicit : bool, optional
            If true, then true Robin BC are applied. Attention:
            if desired, domain must be cropped by manually by twice the Stern
            layer thickness lambda_S. Otherwise, enforces
            constant potential gradient across Stern layer region of thickness
            lambda_S. (default:False)
        """
        self.boundary_conditions = []

        # Potential Dirichlet BC
        self.u0 = self.delta_u_scaled / 2.0
        self.u1 = - self.delta_u_scaled / 2.0

        if implicit: # implicitly treat Stern layer via Robin BC
            self.logger.info('Implicitly treating Stern layer via Robin BC')

            self.logger.info('{:>{lwidth}s} u0 + lambda_S*dudx = {:< 8.4g}'.format(
              'Left hand side Robin boundary condition', self.u0, lwidth=self.label_width))
            self.logger.info('{:>{lwidth}s} u1 + lambda_S*dudx = {:< 8.4g}'.format(
              'Right hand side Robin boundary condition', self.u1, lwidth=self.label_width))
            self.boundary_conditions.extend([
              lambda x: self.leftPotentialRobinBC(x,self.lambda_S_scaled,self.u0),
              lambda x: self.rightPotentialRobinBC(x,self.lambda_S_scaled,self.u1) ])

        else: # explicitly treat Stern layer via linear regime
            self.logger.info('Explicitly treating Stern layer as uniformly charged regions')

            # set left and right hand side outer Helmholtz plane
            self.lhs_ohp = self.x0_scaled + self.lambda_S_scaled
            self.rhs_ohp = self.x0_scaled + self.L_scaled - self.lambda_S_scaled

            self.logger.info('{:>{lwidth}s} u0 = {:< 8.4g}'.format(
              'Left hand side Dirichlet boundary condition', self.u0, lwidth=self.label_width))
            self.logger.info('{:>{lwidth}s} u1 = {:< 8.4g}'.format(
              'Right hand side Dirichlet boundary condition', self.u1, lwidth=self.label_width))
            self.boundary_conditions.extend([
              lambda x: self.leftPotentialDirichletBC(x,self.u0),
              lambda x: self.rightPotentialDirichletBC(x,self.u1) ])


        N0 = self.L_scaled*self.c_scaled # total amount of species in cell
        for k in range(self.M):
          self.logger.info('{:>{lwidth}s} j0 = {:<8.4g}'.format(
            'Ion species {:02d} left hand side concentration Flux boundary condition'.format(k),
            0.0, lwidth=self.label_width))
          self.logger.info('{:>{lwidth}s} N0 = {:<8.4g}'.format(
            'Ion species {:02d} number conservation constraint'.format(k),
            N0[k], lwidth=self.label_width))

          self.boundary_conditions.extend(  [
            lambda x, k=k: self.leftControlledVolumeSchemeFluxBC(x,k),
            lambda x, k=k, N0=N0[k]: self.numberConservationConstraint(x,k,N0) ] )

    # TODO: meaningful test for Dirichlet BC
    def useStandardDirichletBC(self):
        """Dirichlet BC for all variables at all boundaries"""
        self.boundary_conditions = []

        self.u0 = self.delta_u_scaled
        self.u1 = 0

        self.logger.info('Left hand side potential Dirichlet boundary condition:                     u0 = {:> 8.4g}'.format(self.u0))
        self.logger.info('Right hand side potential Dirichlet boundary condition:                    u1 = {:> 8.4g}'.format(self.u1))

        # set up boundary conditions
        self.boundary_conditions.extend( [
          lambda x: self.leftPotentialDirichletBC(x,self.u0),
          lambda x: self.rightPotentialDirichletBC(x,self.u1) ] )

        for k in range(self.M):
          self.logger.info('Ion species {:02d} left hand side concentration Dirichlet boundary condition:  c0 = {:> 8.4g}'.format(k,self.c_scaled[k]))
          self.logger.info('Ion species {:02d} right hand side concentration Dirichlet boundary condition: c1 = {:> 8.4g}'.format(k,self.c_scaled[k]))
          self.boundary_conditions.extend( [
            lambda x, k=k: self.leftDirichletBC(x,k,self.c_scaled[k]),
            lambda x, k=k: self.rightDirichletBC(x,k,self.c_scaled[k]) ] )

    # boundary conditions and constraints building blocks:
    def leftFiniteDifferenceSchemeFluxBC(self,x,k,j0=0):
        """
        Parameters
        ----------
        x : (Ni,) ndarray
            N-valued variable vector
        k : int
            ion species (-1 for potential)
        j0 : float
            flux of ion species `k` at left hand boundary

        Returns
        -------
        float: boundary condition residual
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        # 2nd order right hand side finite difference scheme:
        # df0dx = 1 / (2*dx) * (-3 f0 + 4 f1 - f2 ) + O(dx^2)
        # - dndx - z n dudx = j0
        dndx = -3.0*nijk[0] + 4.0*nijk[1] - nijk[2]
        dudx = -3.0*uij[0]  + 4.0*uij[1]  - uij[2]
        bcval = - dndx - self.zi0[k,0]*nijk[0]*dudx - 2.0*self.dx*j0

        self.logger.debug(
            'Flux BC F[0]  = - dndx - z n dudx - 2*dx*j0 = {:> 8.4g}'.format(bcval))
        self.logger.debug(
            '   = - ({:.2f}) - ({:.0f})*{:.2f}*({:.2f}) - 2*{:.2f}*({:.2f})'.format(
                dndx, self.zi0[k,0], nijk[0], dudx, self.dx, j0))
        return bcval

    def rightFiniteDifferenceSchemeFluxBC(self,x,k,j0=0):
        """
        See ```leftFiniteDifferenceSchemeFluxBC```
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        # 2nd order left hand side finite difference scheme:
        # df0dx = 1 / (2*dx) * (-3 f0 + 4 f1 - f2 ) + O(dx^2)
        # - dndx - z n dudx = j0
        dndx = 3.0*nijk[-1] - 4.0*nijk[-2] + nijk[-3]
        dudx = 3.0*uij[-1]  - 4.0*uij[-2]  + uij[-3]
        bcval = - dndx - self.zi0[k,-1]*nijk[-1]*dudx - 2.0*self.dx*j0

        self.logger.debug(
            'FD flux BC F[-1]  = - dndx - z n dudx - 2*dx*j0 = {:> 8.4g}'.format(bcval))
        self.logger.debug(
            '  = - {:.2f} - {:.0f}*{:.2f}*{:.2f} - 2*{:.2f}*{:.2f}'.format(
                dndx, self.zi0[k,-1], nijk[-1], dudx, self.dx, j0))
        return bcval

    def leftControlledVolumeSchemeFluxBC(self,x,k,j0=0):
        """
        Compute left hand side flux boundary condition residual in accord with
        controlled volume scheme.

        Parameters
        ----------
        x : (Ni,) ndarray
            N-valued variable vector
        k : int
            ion species (-1 for potential)
        j0 : float
            flux of ion species `k` at left hand boundary

        Returns
        -------
        float: boundary condition residual
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]

        # flux by controlled volume scheme:

        bcval = ( + B(self.z[k]*(uij[0]-uij[1]))*nijk[1]
                  - B(self.z[k]*(uij[1]-uij[0]))*nijk[0] - self.dx*j0 )

        self.logger.debug(
            'CV flux BC F[0]  = n1*B(z(u0-u1)) - n0*B(z(u1-u0)) - j0*dx = {:> 8.4g}'.format(bcval))
        return bcval

    def rightControlledVolumeSchemeFluxBC(self,x,k,j0=0):
        """
        Compute right hand side flux boundary condition residual in accord with
        controlled volume scheme. See ``leftControlledVolumeSchemeFluxBC``
        """
        uij = x[:self.Ni]
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]

        # flux by controlled volume scheme:

        bcval = ( + B(self.z[k]*(uij[-2]-uij[-1]))*nijk[-1]
                  - B(self.z[k]*(uij[-1]-uij[-2]))*nijk[-2] - self.dx*j0 )

        self.logger.debug(
            'CV flux BC F[-1]  = n[-1]*B(z(u[-2]-u[-1])) - n[-2]*B(z(u[-1]-u[-2])) - j0*dx = {:> 8.4g}'.format(bcval))
        return bcval

    def leftPotentialDirichletBC(self,x,u0=0):
        return self.leftDirichletBC(x,-1,u0)

    def leftDirichletBC(self,x,k,x0=0):
        """Construct Dirichlet BC at left boundary"""
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[0] - x0

    def rightPotentialDirichletBC(self,x,x0=0):
        return self.rightDirichletBC(x,-1,x0)

    def rightDirichletBC(self,x,k,x0=0):
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[-1] - x0

    def leftPotentialRobinBC(self,x,lam,u0=0):
        return self.leftRobinBC(x,-1,lam,u0)

    def leftRobinBC(self,x,k,lam,x0=0):
        """
        Compute left hand side Robin (u + lam*dudx = u0 ) BC at in accord with
        2nd order finite difference scheme.

        Parameters
        ----------
        x : (Ni,) ndarray
            N-valued variable vector
        k : int
            ion species (-1 for potential)
        lam: float
            BC coefficient, corresponds to Stern layer thickness
            if applied to potential variable in PNP problem. Here, this steric
            layer is assumed to constitute a region of uniform charge density
            and thus linear potential drop across the interface.
        x0 : float
            right hand side value of BC, corresponds to potential beyond Stern
            layer if applied to poential variable in PNP system.

        Returns
        -------
        float: boundary condition residual
        """
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[0] + lam/(2*self.dx)* ( 3.0*nijk[0] - 4.0*nijk[1] + nijk[2] ) - x0

    def rightPotentialRobinBC(self,x,lam,u0=0):
        return self.rightRobinBC(x,-1,lam,u0)

    def rightRobinBC(self,x,k,lam,x0=0):
        """Construct Robin (u + lam*dudx = u0 ) BC at right boundary."""
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]
        return nijk[-1] + lam/(2*self.dx) * ( 3.0*nijk[-1] - 4.0*nijk[-2] + nijk[-3] ) - x0

    def numberConservationConstraint(self,x,k,N0):
        """N0: total amount of species, k: ion species"""
        nijk = x[(k+1)*self.Ni:(k+2)*self.Ni]

        ## TODO: this integration scheme assumes constant concentrations within
        ## an interval. Adapt to controlled volume scheme!

        # rescale to fit interval
        N = np.sum(nijk*self.dx) * self.N / self.Ni
        constraint_val = N - N0

        self.logger.debug(
            'Number conservation constraint F(x)  = N - N0 = {:.4g} - {:.4g} = {:.4g}'.format(
                N, N0, constraint_val ) )
        return constraint_val

    # TODO: remove or standardize
    # def leftNeumannBC(self,x,j0):
    #   """Construct finite difference Neumann BC (flux BC) at left boundary"""
    #   # right hand side first derivative of second order error
    #   # df0dx = 1 / (2*dx) * (-3 f0 + 4 f1 - f2 ) + O(dx^2) = j0
    #   bcval = -3.0*x[0] + 4.0*x[1] - x[2] - 2.0*self.dx*j0
    #   self.logger.debug(
    #     'Neumann BC F[0]  = -3*x[0]  + 4*x[1]  - x[2]  = {:> 8.4g}'.format(bcval))
    #   return bcval
    #
    # def rightNeumannBC(self,x,j0):
    #   """Construct finite difference Neumann BC (flux BC) at right boundray"""
    #   # left hand side first derivative of second order error
    #   # dfndx = 1 / (2*dx) * (+3 fn - 4 fn-1 + fn-2 ) + O(dx^2) = 0
    #   bcval = 3.0*x[-1] - 4.0*x[-2] + x[-3] - 2.0*self.dx*j0
    #   self.logger.debug(
    #     'Neumann BC F[-1] = -3*x[-1] + 4*x[-2] - nijk[-3] = {:> 8.4g}'.format(bcval))
    #   return bcval

    # standard Poisson equation residual for potential
    def poisson_pde(self,x):
        """Returns Poisson equation resiudal by applying 2nd order FD scheme"""
        uij1 = x[:self.Ni]
        self.logger.debug(
          'potential range [u_min, u_max] = [ {:>.4g}, {:>.4g} ]'.format(
            np.min(uij1),np.max(uij1)))

        nij1 = x[self.Ni:(self.M+1)*self.Ni]

        nijk1 = nij1.reshape( self.M, self.Ni )
        for k in range(self.M):
          self.logger.debug(
            'ion species {:02d} concentration range [c_min, c_max] = [ {:>.4g}, {:>.4g} ]'.format(
              k,np.min(nijk1[k,:]),np.max(nijk1[k,:])))

        # M rows (ion species), N_i cols (grid points)
        zi0nijk1 = self.zi0*nijk1 # z_ik*n_ijk
        for k in range(self.M):
          self.logger.debug(
            'ion species {:02d} charge range [z*c_min, z*c_max] = [ {:>.4g}, {:>.4g} ]'.format(
              k,np.min(zi0nijk1[k,:]), np.max(zi0nijk1[k,:])))

        # charge density sum_k=1^M (z_ik*n_ijk)
        rhoij1 = zi0nijk1.sum(axis=0)
        self.logger.debug(
          'charge density range [rho_min, rho_max] = [ {:>.4g}, {:>.4g} ]'.format(
            np.min(rhoij1),np.max(rhoij1)))

        # reduced Poisson equation: d2udx2 = rho
        Fu = -(np.roll(uij1, -1)-2*uij1+np.roll(uij1, 1))-0.5*rhoij1*self.dx**2

        # linear potential regime due to steric effects incorporated here
        # TODO: incorporate "spatially finite" BC into Robin BC functions
        # replace left and right hand side residuals with linear potential FD

        if not np.isnan(self.lhs_ohp):
            lhs_linear_regime_ndx = (self.X <= self.lhs_ohp)
            lhs_ohp_ndx = np.max( np.nonzero( lhs_linear_regime_ndx ) )

            self.logger.debug(
              'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled <= {:>.4g}'.format(
                np.count_nonzero(lhs_linear_regime_ndx), lhs_ohp_ndx, self.lhs_ohp) )

            # dudx = (u[ohp]-u[0])/lambda_S within Stern layer
            Fu[lhs_linear_regime_ndx] = (
                ( np.roll(uij1,-1) - uij1 )[lhs_linear_regime_ndx]
                    * self.lambda_S_scaled - (uij1[lhs_ohp_ndx]-uij1[0])*self.dx )

        if not np.isnan(self.rhs_ohp):
            rhs_linear_regime_ndx = (self.X >= self.rhs_ohp)
            rhs_ohp_ndx = np.min( np.nonzero( rhs_linear_regime_ndx ) )

            self.logger.debug(
              'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled >= {:>.4g}'.format(
                 np.count_nonzero(rhs_linear_regime_ndx), rhs_ohp_ndx, self.rhs_ohp) )

            # dudx = (u[ohp]-u[0])/lambda_S within Stern layer
            Fu[rhs_linear_regime_ndx] = (
                ( uij1 - np.roll(uij1,1) )[rhs_linear_regime_ndx]
                    * self.lambda_S_scaled - (uij1[-1]-uij1[rhs_ohp_ndx])*self.dx )

        Fu[0] = self.boundary_conditions[0](x)
        Fu[-1] = self.boundary_conditions[1](x)

        self.logger.debug('Potential BC residual Fu[0]  = {:> 8.4g}'.format(Fu[0]))
        self.logger.debug('Potential BC residual Fu[-1] = {:> 8.4g}'.format(Fu[-1]))

        return Fu

    def nernst_planck_pde(self,x):
        """Returns Nernst-Planck equation resiudal by applying controlled
        volume scheme"""
        uij1 = x[:self.Ni]

        self.logger.debug(
          'potential range [u_min, u_max] = [ {:>.4g}, {:>.4g} ]'.format(
            np.min(uij1),np.max(uij1)))

        nij1 = x[self.Ni:(self.M+1)*self.Ni]
        nijk1 = nij1.reshape( self.M, self.Ni )
        for k in range(self.M):
          self.logger.debug(
            'ion species {:02d} concentration range [c_min, c_max] = [ {:>.4g}, {:>.4g} ]'.format(
              k,np.min(nijk1[k,:]),np.max(nijk1[k,:]) ) )

        Fn = np.zeros([self.M, self.Ni])
        # loop over k = 1..M reduced Nernst-Planck equations:
        # - d2nkdx2 - ddx (zk nk dudx ) = 0
        for k in range(self.M):
          # conrolled volume implementation: constant flux across domain
          Fn[k,:] = (
            + B(self.zi0[k,:]*(uij1 - np.roll(uij1,-1))) * np.roll(nijk1[k,:],-1)
            - B(self.zi0[k,:]*(np.roll(uij1,-1) - uij1)) * nijk1[k,:]
            - B(self.zi0[k,:]*(np.roll(uij1,+1) - uij1)) * nijk1[k,:]
            + B(self.zi0[k,:]*(uij1 - np.roll(uij1,+1))) * np.roll(nijk1[k,:],+1) )

          # controlled volume implementation: flux j = 0 in every grid point
          #
          # Fn[k,:] =  (
          #   B(self.zi0[k,:]*(uij1 - np.roll(uij1,-1)))*np.roll(nijk1[k,:],-1)
          #   - B(self.zi0[k,:]*(np.roll(uij1,-1) - uij1))*nijk1[k,:] )

          # linear potential regime due to steric effects incorporated here
          # TODO: incorporate "spatially finite" BC into Robin BC functions
          # replace left and right hand side residuals with linear potential FD

          # left and right hand side outer Helmholtz plane
          # lhs_ohp = self.x0_scaled + self.lambda_S_scaled
          # rhs_ohp = self.x0_scaled + self.L_scaled - self.lambda_S_scaled
          #
          # lhs_linear_regime_ndx = (self.X <= lhs_ohp)
          # rhs_linear_regime_ndx = (self.X >= rhs_ohp)
          #
          # lhs_ohp_ndx = np.max( np.nonzero( lhs_linear_regime_ndx ) )
          # rhs_ohp_ndx = np.min( np.nonzero( rhs_linear_regime_ndx ) )
          #
          # self.logger.debug(
          #   'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled <= {:>.4g}'.format(
          #     np.count_nonzero(lhs_linear_regime_ndx), lhs_ohp_ndx, lhs_ohp) )
          # self.logger.debug(
          #   'selected {:d} grid points within lhs OHP at grid point index {:d} with x_scaled >= {:>.4g}'.format(
          #      np.count_nonzero(rhs_linear_regime_ndx), rhs_ohp_ndx, rhs_ohp) )
          #
          # # zero concentration gradient in Stern layer
          # Fn[k,lhs_linear_regime_ndx] = (
          #     ( np.roll(nijk1[k,:],-1)-np.roll(nijk1[k,:],1))[lhs_linear_regime_ndx])
          # Fn[k,rhs_linear_regime_ndx] = (
          #     ( np.roll(nijk1[k,:],-1)-np.roll(nijk1[k,:],1))[rhs_linear_regime_ndx])

          Fn[k,0]  = self.boundary_conditions[2*k+2](x)
          Fn[k,-1] = self.boundary_conditions[2*k+3](x)

          self.logger.debug(
            'ion species {k:02d} BC residual Fn[{k:d},0]  = {:> 8.4g}'.format(
              Fn[k,0],k=k))
          self.logger.debug(
            'ion species {k:02d} BC residual Fn[{k:d},-1]  = {:> 8.4g}'.format(
              Fn[k,-1],k=k))

        return Fn

    # non-linear system, "controlled volume" method
    # Selbherr, S. Analysis and Simulation of Semiconductor Devices, Spriger 1984
    def G(self, x):
        """Non-linear system

        Discretization of Poisson-Nernst-Planck system with M ion species.
        Implements "controlled volume" method as found in

          Selbherr, Analysis and Simulation of Semiconductor Devices, Spriger 1984

        Parameters
        ----------
        x : ((M+1)*Ni,) ndarray
            system variables. 1D array of (M+1)*Ni values, wher M is number of
            ion sepcies, Ni number of spatial discretization points. First Ni
            entries are expected to contain potential, following M*Ni points
            contain ion concentrations.
        Returns
        --------
        residual: ((M+1)*Ni,) ndarray
        """
        # reduced Poisson equation: d2udx2 = rho
        Fu = self.potential_pde(x)
        Fn = self.concentration_pde(x)

        # Apply constraints if set (not implemented properly, do not use):
        if len(self.g) > 0:
            Flam = np.array([g(x) for g in self.g])
            F = np.concatenate([Fu,Fn.flatten(),Flam])
        else:
            F = np.concatenate([Fu,Fn.flatten()])

        return F

    @property
    def I(self): # ionic strength
        """Compute the system's ionic strength from charges and concentrations.

        Returns
        -------
        I : float
            ionic strength ( 1/2 * sum(z_i^2*c_i) )
            [concentration unit, i.e. mol m^-3]
        """
        return 0.5*np.sum( np.square(self.z) * self.c )

    @property
    def lambda_D(self):
        """Compute the system's Debye length.

        Returns
        -------
        lambda_D : float
            Debye length, sqrt( epsR*eps*R*T/(2*F^2*I) ) [length unit, i.e. m]
        """
        return np.sqrt(
            self.relative_permittivity*self.vacuum_permittivity*self.R*self.T/(
                2.0*self.F**2*self.I ) )

    # default 0.1 mM (i.e. mol/m^3) NaCl aqueous solution
    def init(self,
        c = np.array([0.1,0.1]),
        z = np.array([1,-1]),
        L = 100e-9, # 100 nm
        lambda_S=0, # Stern layer (compact layer) thickness
        x0 = 0, # zero position
        T = 298.15,
        delta_u = 0.05, # potential difference [V]
        relative_permittivity = 79,
        vacuum_permittivity   = sc.epsilon_0,
        R = sc.value('molar gas constant'),
        F = sc.value('Faraday constant'),
        N = 200, # number of grid segments, number of grid points Ni = N + 1
        e = 1e-10, # absolute tolerance, TODO: switch to standaradized measure
        maxit = 20, # maximum number of Newton iterations
        solver = None,
        options = None,
        potential0 = None,
        concentration0 = None ):
        """Initializes a 1D Poisson-Nernst-Planck system description.

        Expects quantities in SI units per default.

        Parameters
        ----------
        c : (M,) ndarray, optional
            bulk concentrations of each ionic species [mol/m^3]
            (default: [ 0.1, 0.1 ])
        z : (M,) ndarray, optional
            charge of each ionic species [1] (default: [ +1, -1 ])
        x0 : float, optional
            left hand side reference position (default: 0)
        L : float, optional
            1D domain size [m] (default: 100e-9)
        lambda_S: float, optional
            Stern layer thickness in case of Robin BC [m] (default: 0)
        T : float, optional
            temperature of the solution [K] (default: 298.15)
        delta_u : float, optional
            potential drop across 1D cell [V] (default: 0.05)
        relative_permittivity: float, optional
            relative permittivity of the ionic solution [1] (default: 79)
        vacuum_permittivity: float, optional
            vacuum permittivity [F m^-1] (default: 8.854187817620389e-12 )
        R : float, optional
            molar gas constant [J mol^-1 K^-1] (default: 8.3144598)
        F : float, optional
            Faraday constant [C mol^-1] (default: 96485.33289)
        N : int, optional
            number of discretization grid segments (default: 200)
        e : float, optional
            absolute tolerance for Newton solver convergence (default: 1e-10)
        maxit : int, optional
            maximum number of Newton iterations (default: 20)
        solver: func( funx(x), x0), optional
            solver to use (default: None, will use own simple Newton solver)
        potential0: (N+1,) ndarray, optional (default: None)
            potential initial values
        concentration0: (M,N+1) ndarray, optional (default: None)
            concentration initial values
        """

        self.logger = logging.getLogger(__name__)

        assert len(c) == len(z), "Provide concentration AND charge for ALL ion species!"

        # TODO: integrate with constructor initialization parameters above
        # default solver settings
        self.converged = False # solver's convergence flag
        self.N      = N     # discretization segments
        self.e      = e     # Newton solver default tolerance
        self.maxit  = maxit # Newton solver maximum iterations

        # default output settings
        # self.output = False   # let Newton solver output convergence plots...
        # self.outfreq = 1      # ...at every nth iteration
        self.label_width = 40 # charcater width of quantity labels in log

        # standard governing equations
        self.potential_pde      = self.poisson_pde
        self.concentration_pde  = self.nernst_planck_pde

        # empty BC
        self.boundary_conditions = []
        # empty constraints
        self.g = [] # list of constrain functions, not fully implemented / tested

        # system parameters
        self.M = len(c) # number of ion species

        self.c  = c # concentrations
        self.z  = z # number charges
        self.T  = T # temperature
        self.L  = L # 1d domain size
        self.lambda_S = lambda_S # Stern layer thickness
        self.x0 = x0 # reference position
        self.delta_u = delta_u # potential difference


        self.relative_permittivity = relative_permittivity
        self.vacuum_permittivity   = vacuum_permittivity
        # R = N_A * k_B
        # (universal gas constant = Avogadro constant * Boltzmann constant)
        self.R                     = R
        self.F                     = F

        self.f                     = F / (R*T)  # for convenience

        # print all quantities to log
        for i, (c, z) in enumerate(zip(self.c, self.z)):
            self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
                "ion species {:02d} concentration c".format(i), c, lwidth=self.label_width))
            self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
                "ion species {:02d} number charge z".format(i), z, lwidth=self.label_width))

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'temperature T', self.T, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'domain size L', self.L, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'compact layer thickness lambda_S', self.lambda_S, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'reference position x0', self.x0, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'potential difference delta_u', self.delta_u, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'relative permittivity eps_R', self.relative_permittivity, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'vacuum permittivity eps_0', self.vacuum_permittivity, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'universal gas constant R', self.R, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'Faraday constant F', self.F, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'f = F / (RT)', self.f, lwidth=self.label_width))

        # scaled units for dimensionless formulation

        # length unit chosen as Debye length lambda
        self.l_unit = self.lambda_D

        # concentration unit is ionic strength
        self.c_unit = self.I

        # no time unit for now, only steady state
        # self.t_unit = self.l_unit**2 / self.Dn # fixes Dn_scaled = 1

        self.u_unit = self.R * self.T / self.F  # thermal voltage

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'spatial unit [l]', self.l_unit, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'concentration unit [c]', self.c_unit, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'potential unit [u]', self.u_unit, lwidth=self.label_width))

        # domain
        self.L_scaled = self.L / self.l_unit

        # compact layer
        self.lambda_S_scaled = self.lambda_S / self.l_unit

        # reference position
        self.x0_scaled = self.x0 / self.l_unit

        # bulk conectrations
        self.c_scaled = self.c / self.c_unit

        # potential difference
        self.delta_u_scaled = self.delta_u / self.u_unit

        # print scaled quantities to log
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'reduced domain size L*', self.L_scaled, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'reduced compact layer thickness lambda_S*', self.lambda_S_scaled, lwidth=self.label_width))
        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'reduced reference position x0*', self.x0_scaled, lwidth=self.label_width))

        for i, c_scaled in enumerate(self.c_scaled):
            self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
                "ion species {:02d} reduced concentration c*".format(i),
                c_scaled, lwidth=self.label_width))

        self.logger.info('{:<{lwidth}s} {:> 8.4g}'.format(
            'reduced potential delta_u*', self.delta_u_scaled, lwidth=self.label_width))

        # per default, no outer Helmholtz plane
        self.lhs_ohp = np.nan
        self.rhs_ohp = np.nan

        # self.xi0 = None
        # initialize initial value arrays
        if potential0 is not None:
            self.ui0 = potential0 / self.u_unit
        else:
            self.ui0 = None

        if concentration0 is not None:
            self.ni0 = concentration0 / self.c_unit
        else:
            self.ni0 = None

        self.zi0 = None

    def __init__(self, *args, **kwargs):
        """Constructor, see init doc string for arguments.

        Additional Parameters
        ---------------------
        solver: str or func (default: None)
            solver to use. If str, then selected from scipy optimizers.
        options: dict, optional (default: None)
            options object for scipy solver
        """
        self.init(*args, **kwargs)

        if 'solver' in kwargs:
            self.solver = kwargs['solver']
        else:
            self.solver = self.newton

        if 'options' in kwargs:
            self.options = kwargs['options']
        else:
            self.options = None

        self.discretize()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#                  Lukas Elflein, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""
Electrochemistry module utility functions

Copyright 2019, 2020 IMTEK Simulation
University of Freiburg

Authors:

  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
  Lukas Elflein <elfleinl@cs.uni-freiburg.de>
"""

import matplotlib.pyplot as plt

def get_centers(bins):
    """Return the center of the provided bins.

    Example:
    >>> get_centers(bins=np.array([0.0, 1.0, 2.0]))
    array([ 0.5,  1.5])
    """
    bins = bins.astype(float)
    return (bins[:-1] + bins[1:]) / 2


def plot_dist(histogram, name, reference_distribution=None):
    """Plot histogram with an optional reference distribution."""
    hist, bins = histogram
    width = 1 * (bins[1] - bins[0])
    centers = get_centers(bins)

    fi, ax = plt.subplots()
    ax.bar( centers, hist, align='center', width=width, label='Empirical distribution',
            edgecolor="none")

    if reference_distribution is not None:
        ref = reference_distribution(centers)
        ref /= sum(ref)
        ax.plot(centers, ref, color='red', label='Target distribution')

    plt.title(name)
    plt.legend()
    plt.xlabel('Distance ' + name)
    plt.savefig(name + '.png')



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2018) James Kermode, King's College London
#                       Lars Pastewka, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

"""
Simple pair potential.
"""

#
# Coding convention
# * All numpy arrays are suffixed with the array dimensions
# * The suffix stands for a certain type of dimension:
#   - n: Atomic index, i.e. array dimension of length nb_atoms
#   - p: Pair index, i.e. array dimension of length nb_pairs
#   - c: Cartesian index, array dimension of length 3
#

import numpy as np

from scipy.sparse import bsr_matrix, vstack, hstack

import ase

from ...neighbours import neighbour_list, first_neighbours
from ..calculator import MatscipyCalculator
from ...numpy_tricks import mabincount


###

class LennardJonesCut():
    """
    Functional form for a 12-6 Lennard-Jones potential with a hard cutoff.
    Energy is shifted to zero at cutoff.
    """

    def __init__(self, epsilon, sigma, cutoff):
        self.epsilon = epsilon
        self.sigma = sigma
        self.cutoff = cutoff
        self.offset = (sigma/cutoff)**12 - (sigma/cutoff)**6

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((r6-1) * r6 - self.offset)

    def get_cutoff(self):
        return self.cutoff

    def first_derivative(self, r):
        r = (self.sigma / r)
        r6 = r**6
        return -24 * self.epsilon / self.sigma * (2*r6-1) * r6 * r

    def second_derivative(self, r):
        r2 = (self.sigma / r)**2
        r6 = r2**3
        return 24 * self.epsilon/self.sigma**2 * (26*r6-7) * r6 * r2

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class LennardJonesQuadratic():
    """
    Functional form for a 12-6 Lennard-Jones potential with a soft cutoff.
    Energy, its first and second derivative are shifted to zero at cutoff.
    """

    def __init__(self, epsilon, sigma, cutoff):
        self.epsilon = epsilon
        self.sigma = sigma
        self.cutoff = cutoff
        self.offset_energy = (sigma/cutoff)**12 - (sigma/cutoff)**6
        self.offset_force = 6/cutoff * \
            (-2*(sigma/cutoff)**12+(sigma/cutoff)**6)
        self.offset_dforce = (1/cutoff**2) * \
            (156*(sigma/cutoff)**12-42*(sigma/cutoff)**6)

    def get_cutoff(self):
        return self.cutoff

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((r6-1)*r6-self.offset_energy - (r-self.cutoff) * self.offset_force - ((r - self.cutoff)**2/2) * self.offset_dforce)

    def first_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((6/r) * (-2*r6+1) * r6 - self.offset_force - (r-self.cutoff) * self.offset_dforce)

    def second_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((1/r**2) * (156*r6-42) * r6 - self.offset_dforce)

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class LennardJonesLinear():
    """
    Function form of a 12-6 Lennard-Jones potential with a soft cutoff
    The energy and the force are shifted at the cutoff.
    """

    def __init__(self, epsilon, sigma, cutoff):
        self.epsilon = epsilon
        self.sigma = sigma
        self.cutoff = cutoff
        self.offset_energy = (sigma/cutoff)**12 - (sigma/cutoff)**6
        self.offset_force = 6/cutoff * \
            (-2*(sigma/cutoff)**12+(sigma/cutoff)**6)

    def get_cutoff(self):
        return self.cutoff

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((r6-1) * r6 - self.offset_energy - (r-self.cutoff) * self.offset_force)

    def first_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((6/r) * (-2*r6+1) * r6 - self.offset_force)

    def second_derivative(self, r):
        r6 = (self.sigma / r)**6
        return 4 * self.epsilon * ((1/r**2) * (156*r6-42) * r6)

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))


###

class FeneLJCut():
    """
    Finite extensible nonlinear elastic(FENE) potential for a bead-spring polymer model.
    For the Lennard-Jones interaction a LJ-cut potential is used. Due to choice of the cutoff (rc=2^(1/6) sigma)
    it ensures a continous potential and force at the cutoff.
    """

    def __init__(self, K, R0, epsilon, sigma):
        self.K = K
        self.R0 = R0
        self.epsilon = epsilon
        self.sigma = sigma

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r6 = (self.sigma/r)**6
        bond = -0.5 * self.K * self.R0**2 * np.log(1-(r/self.R0)**2)
        lj = 4 * self.epsilon * (r6-1) * r6 + self.epsilon
        return bond + lj

    def first_derivative(self, r):
        r6 = (self.sigma/r)**6
        bond = self.K * r / (1-(r/self.R0)**2)
        lj = -24 * self.epsilon * (2*r6/r-1/r) * r6
        return bond + lj

    def second_derivative(self, r):
        r6 = (self.sigma/r)**6
        invLength = 1 / (1-(r/self.R0)**2)
        bond = K * invLength + 2 * K * r**2 * invLength**2 / self.R0**2
        lj = 4 * self.epsilon * ((1/r**2) * (156*r6-42) * r6)
        return bond + lj

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))


###

class LennardJones84():
    """
    Function form of a 8-4 Lennard-Jones potential, used to model the structure of a CuZr.
    Kobayashi, Shinji et. al. "Computer simulation of atomic structure of Cu57Zr43 amorphous alloy."
    Journal of the Physical Society of Japan 48.4 (1980): 1147-1152.
    """

    def __init__(self, C1, C2, C3, C4, cutoff):
        self.C1 = C1
        self.C2 = C2
        self.C3 = C3
        self.C4 = C4
        self.cutoff = cutoff

    def get_cutoff():
        return self.cutoff

    def __call__(self, r):
        """
        Return function value (potential energy).
        """
        r4 = (1 / r)**4
        return (self.C2*r4-self.C1) * r4 + self.C3 * r + self.C4

    def first_derivative(self, r):
        r4 = (1 / r)**4
        return (-8 * self.C2*r4/r+4*self.C1/r) * r4 + self.C3

    def second_derivative(self, r):
        r4 = (1 / r)**4
        return (72 * self.C2 * r4 / r**2 - 20 * self.C1 / r**2) * r4

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class PairPotential(MatscipyCalculator):
    implemented_properties = ['energy', 'free_energy', 'stress', 'forces', 'hessian']
    default_parameters = {}
    name = 'PairPotential'

    def __init__(self, f, cutoff=None):
        MatscipyCalculator.__init__(self)
        self.f = f

        self.dict = {x: obj.get_cutoff() for x, obj in f.items()}
        self.df = {x: obj.derivative(1) for x, obj in f.items()}
        self.df2 = {x: obj.derivative(2) for x, obj in f.items()}

    def calculate(self, atoms, properties, system_changes):
        MatscipyCalculator.calculate(self, atoms, properties, system_changes)

        nb_atoms = len(self.atoms)
        atnums = self.atoms.numbers
        atnums_in_system = set(atnums)

        i_p, j_p, r_p, r_pc = neighbour_list('ijdD', self.atoms, self.dict)

        e_p = np.zeros_like(r_p)
        de_p = np.zeros_like(r_p)
        for params, pair in enumerate(self.dict):
            if pair[0] == pair[1]:
                mask1 = atnums[i_p] == pair[0]
                mask2 = atnums[j_p] == pair[0]
                mask = np.logical_and(mask1, mask2)

                e_p[mask] = self.f[pair](r_p[mask])
                de_p[mask] = self.df[pair](r_p[mask])

            if pair[0] != pair[1]:
                mask1 = np.logical_and(
                    atnums[i_p] == pair[0], atnums[j_p] == pair[1])
                mask2 = np.logical_and(
                    atnums[i_p] == pair[1], atnums[j_p] == pair[0])
                mask = np.logical_or(mask1, mask2)

                e_p[mask] = self.f[pair](r_p[mask])
                de_p[mask] = self.df[pair](r_p[mask])

        epot = 0.5*np.sum(e_p)

        # Forces
        df_pc = -0.5*de_p.reshape(-1, 1)*r_pc/r_p.reshape(-1, 1)

        f_nc = mabincount(j_p, df_pc, nb_atoms) - mabincount(i_p, df_pc, nb_atoms)

        # Virial
        virial_v = -np.array([r_pc[:, 0] * df_pc[:, 0],               # xx
                              r_pc[:, 1] * df_pc[:, 1],               # yy
                              r_pc[:, 2] * df_pc[:, 2],               # zz
                              r_pc[:, 1] * df_pc[:, 2],               # yz
                              r_pc[:, 0] * df_pc[:, 2],               # xz
                              r_pc[:, 0] * df_pc[:, 1]]).sum(axis=1)  # xy

        self.results = {'energy': epot,
                        'free_energy': epot,
                        'stress': virial_v/self.atoms.get_volume(),
                        'forces': f_nc}

    ###

    def get_hessian(self, atoms, format='dense', divide_by_masses=False):
        """
        Calculate the Hessian matrix for a pair potential.
        For an atomic configuration with N atoms in d dimensions the hessian matrix is a symmetric, hermitian matrix
        with a shape of (d*N,d*N). The matrix is in general a sparse matrix, which consists of dense blocks of
        shape (d,d), which are the mixed second derivatives. The result of the derivation for a pair potential can be
        found e.g. in:
        L. Pastewka et. al. "Seamless elastic boundaries for atomistic calculations", Phys. Rev. B 86, 075459 (2012).

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        format: "sparse" or "neighbour-list"
            Output format of the hessian matrix.

        divide_by_masses: bool
            if true return the dynamic matrix else hessian matrix 

        Restrictions
        ----------
        This method is currently only implemented for three dimensional systems
        """
        if self.atoms is None:
            self.atoms = atoms

        f = self.f
        dict = self.dict
        df = self.df
        df2 = self.df2

        nb_atoms = len(atoms)
        atnums = atoms.numbers

        i_p, j_p,  r_p, r_pc = neighbour_list('ijdD', atoms, dict)
        first_i = first_neighbours(nb_atoms, i_p)

        e_p = np.zeros_like(r_p)
        de_p = np.zeros_like(r_p)
        dde_p = np.zeros_like(r_p)
        for params, pair in enumerate(dict):
            if pair[0] == pair[1]:
                mask1 = atnums[i_p] == pair[0]
                mask2 = atnums[j_p] == pair[0]
                mask = np.logical_and(mask1, mask2)

                e_p[mask] = f[pair](r_p[mask])
                de_p[mask] = df[pair](r_p[mask])
                dde_p[mask] = df2[pair](r_p[mask])

            if pair[0] != pair[1]:
                mask1 = np.logical_and(
                    atnums[i_p] == pair[0], atnums[j_p] == pair[1])
                mask2 = np.logical_and(
                    atnums[i_p] == pair[1], atnums[j_p] == pair[0])
                mask = np.logical_or(mask1, mask2)

                e_p[mask] = f[pair](r_p[mask])
                de_p[mask] = df[pair](r_p[mask])
                dde_p[mask] = df2[pair](r_p[mask])
        
        n_pc = (r_pc.T/r_p).T
        H_pcc = -(dde_p * (n_pc.reshape(-1, 3, 1)
                           * n_pc.reshape(-1, 1, 3)).T).T
        H_pcc += -(de_p/r_p * (np.eye(3, dtype=n_pc.dtype)
                                    - (n_pc.reshape(-1, 3, 1) * n_pc.reshape(-1, 1, 3))).T).T

        # Sparse BSR-matrix
        if format == "sparse":
            if divide_by_masses:
                masses_n = self.atoms.get_masses()
                geom_mean_mass_p = np.sqrt(masses_n[i_p]*masses_n[j_p])

            if divide_by_masses:
                H = bsr_matrix(((H_pcc.T/geom_mean_mass_p).T, j_p, first_i), shape=(3*nb_atoms, 3*nb_atoms))

            else: 
                H = bsr_matrix((H_pcc, j_p, first_i), shape=(3*nb_atoms, 3*nb_atoms))

            Hdiag_icc = np.empty((nb_atoms, 3, 3))
            for x in range(3):
                for y in range(3):
                    Hdiag_icc[:, x, y] = - \
                        np.bincount(i_p, weights=H_pcc[:, x, y])

            if divide_by_masses:
                H += bsr_matrix(((Hdiag_icc.T/masses_n).T, np.arange(nb_atoms),
                             np.arange(nb_atoms+1)), shape=(3*nb_atoms, 3*nb_atoms))

            else:
                H += bsr_matrix((Hdiag_icc, np.arange(nb_atoms),
                             np.arange(nb_atoms+1)), shape=(3*nb_atoms, 3*nb_atoms))

            return H

        # Dense matrix format
        elif format == "dense":
            H = np.zeros((3*nb_atoms, 3*nb_atoms))
            for atom in range(len(i_p)):
                H[3*i_p[atom]:3*i_p[atom]+3,
                  3*j_p[atom]:3*j_p[atom]+3] += H_pcc[atom]

            Hdiag_icc = np.empty((nb_atoms, 3, 3))
            for x in range(3):
                for y in range(3):
                    Hdiag_icc[:, x, y] = - \
                        np.bincount(i_p, weights=H_pcc[:, x, y])

            Hdiag_ncc = np.zeros((3*nb_atoms, 3*nb_atoms))
            for atom in range(nb_atoms):
                Hdiag_ncc[3*atom:3*atom+3,
                          3*atom:3*atom+3] += Hdiag_icc[atom]

            H += Hdiag_ncc

            if divide_by_masses:
                masses_p = (self.atoms.get_masses()).repeat(3)
                H /= np.sqrt(masses_p.reshape(-1,1)*masses_p.reshape(1,-1))
                return H

            else:
                return H

        # Neighbour list format
        elif format == "neighbour-list":
            return H_pcc, i_p, j_p, r_pc, r_p



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2018) James Kermode, King's College London
#                       Lars Pastewka, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import numpy as np

#
# Parameter sets
# The '__ref__' dictionary entry is the journal reference
#

Stillinger_Weber_PRB_31_5262_Si = {
    '__ref__':  'F. Stillinger and T. Weber, Phys. Rev. B 31, 5262 (1985)',
    'el':            'Si'            ,
    'epsilon':       2.1683          ,
    'sigma':         2.0951          ,
    'costheta0':     0.333333333333  ,
    'A':             7.049556277     ,
    'B':             0.6022245584    ,
    'p':             4               ,
    'a':             1.80            ,
    'lambda1':       21.0            ,
    'gamma':         1.20            
}

Holland_Marder_PRL_80_746_Si = {
    '__ref__':  'D. Holland and M. Marder, Phys. Rev. Lett. 80, 746 (1998)',
    'el':            'Si'            ,
    'epsilon':       2.1683          ,
    'sigma':         2.0951          ,
    'costheta0':     0.333333333333  ,
    'A':             7.049556277     ,
    'B':             0.6022245584    ,
    'p':             4               ,
    'a':             1.80            ,
    'lambda1':       42.0            ,
    'gamma':         1.20            
}

RLC_Vink_JNCS_282_746_Si = {
    '__ref__':  'RLC Vink et al., J. Non-Cryst. Solids 282 (2001)',
    'el':            'Si'            ,
    'epsilon':       1.64833         ,
    'sigma':         2.0951          ,
    'costheta0':     0.333333333333  ,
    'A':             7.049556277     ,
    'B':             0.6022245584    ,
    'p':             4               ,
    'a':             1.80            ,
    'lambda1':       31.5            ,
    'gamma':         1.20            
}

Russo_PRX_8_021040_Si = {
    '__ref__':  'J. Russo et. al. Phys. Rev. X 8, 021040 (2018)',
    'el':            'Si'            ,
    'epsilon':       2.1683          ,
    'sigma':         2.0951          ,
    'costheta0':     0.333333333333  ,
    'A':             7.049556277     ,
    'B':             0.6022245584    ,
    'p':             4               ,
    'a':             1.80            ,
    'lambda1':       18.75           ,
    'gamma':         1.20            
}

def ab(x):
    """
    Compute absolute value (norm) of an array of vectors
    """
    return np.linalg.norm(x, axis=1)

def StillingerWeber(parameters):
    """
    Implementation of the functional form of the Stillinger-Weber potential.

    Reference
    ------------
    F. Stillinger and T. Weber, Physical review B 31.8 5262 (1985)

    """

    el = parameters['el']
    epsilon = parameters['epsilon']
    sigma = parameters['sigma']
    costheta0 = parameters['costheta0']
    A = parameters['A']
    B = parameters['B']
    p = parameters['p']
    a = parameters['a']
    lambda1 = parameters['lambda1']
    gamma = parameters['gamma']

    fR = lambda r: A * epsilon * (B*np.power(sigma/r, p) - 1) * np.exp(sigma/(r-a*sigma))
    dfR = lambda r: - A*epsilon*B*p/r * np.power(sigma/r, p) * np.exp(sigma/(r-a*sigma)) - sigma/np.power(r-a*sigma, 2)*fR(r)
    ddfR = lambda r: A*B*p*epsilon/r * np.power(sigma/r, p) * np.exp(sigma/(r-a*sigma)) * (sigma/np.power(r-a*sigma, 2) + (p+1)/r) \
                     + 2 * sigma / np.power(r-a*sigma, 3) * fR(r) - sigma / np.power(r-a*sigma, 2) * dfR(r)

    fA = lambda r: np.exp(gamma*sigma/(r-a*sigma))
    dfA = lambda r: - gamma * sigma / np.power(r - a*sigma, 2) * fA(r)
    ddfA = lambda r: 2 * gamma * sigma / np.power(r-a*sigma, 3) * fA(r) - gamma * sigma / np.power(r-a*sigma, 2) * dfA(r)

    hf = lambda rik: np.where(ab(rik)<a*sigma, np.exp(gamma*sigma/(ab(rik)-a*sigma)), 0) 
    d2h = lambda rik: -gamma * sigma / np.power(ab(rik)-a*sigma, 2) * hf(rik)
    d22h = lambda rik: 2 * gamma * sigma / np.power(ab(rik)-a*sigma, 3) * hf(rik) - gamma * sigma / np.power(ab(rik)-a*sigma, 2) * d2h(rik)

    g = lambda cost: np.power(cost + costheta0, 2)
    dg = lambda cost: 2 * (cost + costheta0)
    ddg = lambda cost: 2*cost**0

    def F(r, xi, i, p):
        mask = (r < a*sigma)
        F_n = np.zeros_like(r)
        F_n[mask] = fR(r[mask]) + lambda1 * epsilon * fA(r[mask]) * xi[mask]
        return F_n
    def d1F(r, xi, i, p):
        mask = (r < a*sigma)
        d1F_n = np.zeros_like(r)
        d1F_n[mask] = dfR(r[mask]) + lambda1 * epsilon * xi[mask] * dfA(r[mask])
        return d1F_n
    def d2F(r, xi, i, p):
        mask = (r < a*sigma)
        d2F_n = np.zeros_like(r)
        d2F_n[mask] = lambda1 * epsilon * fA(r[mask])     
        return d2F_n
    def d11F(r, xi, i, p):
        mask = (r < a*sigma)
        d11F_n = np.zeros_like(r)
        d11F_n[mask] = ddfR(r[mask]) + lambda1 * epsilon * xi[mask] * ddfA(r[mask])   
        return d11F_n
    def d22F(r, xi, i, p):
        return np.zeros_like(r)
    def d12F(r, xi, i, p):
        mask = (r < a*sigma)
        d12F_n = np.zeros_like(r)
        d12F_n[mask] = lambda1 * epsilon * dfA(r[mask])
        return d12F_n

    G = lambda rij, rik, i, ij, ik: hf(rik) * g(costh(rij, rik))

    d1G = lambda rij, rik, i, ij, ik: (hf(rik) * Dg1(rij, rik).T).T
    d2G = lambda rij, rik, i, ij, ik: (Dh2(rik).T * g(costh(rij, rik)) + hf(rik) * Dg2(rij, rik).T).T

    Dh2 = lambda rik: (d2h(rik) * rik.T / ab(rik)).T 

    Dg1 = lambda rij, rik: (dg(costh(rij, rik)) * c1(rij, rik).T).T 
    Dg2 = lambda rij, rik: (dg(costh(rij, rik)) * c2(rij, rik).T).T

    d11G = lambda rij, rik, i, ij, ik: \
        ((hf(rik) * Dg11(rij, rik).T).T)

    Dg11 = lambda rij, rik: \
        (ddg(costh(rij, rik)) * (c1(rij, rik).reshape(-1, 3, 1) * c1(rij, rik).reshape(-1, 1, 3)).T
         + dg(costh(rij, rik)) * dc11(rij, rik).T).T

    d22G = lambda rij, rik, i, ij, ik: \
        Dg2(rij, rik).reshape(-1, 3, 1) * Dh2(rik).reshape(-1, 1, 3) + Dh2(rik).reshape(-1, 3, 1) * Dg2(rij, rik).reshape(-1, 1, 3) \
        + ((g(costh(rij, rik)) * Dh22(rij, rik).T).T + (hf(rik) * Dg22(rij, rik).T).T)

    Dg22 = lambda rij, rik: \
        (ddg(costh(rij, rik)) * (c2(rij, rik).reshape(-1, 3, 1) * c2(rij, rik).reshape(-1, 1, 3)).T
        + dg(costh(rij, rik)) * dc22(rij, rik).T).T

    Dh22 = lambda rij, rik: \
        (d22h(rik) * (((rik.reshape(-1, 3, 1) * rik.reshape(-1, 1, 3)).T/ab(rik)**2).T).T \
         + d2h(rik) * ((np.eye(3) - ((rik.reshape(-1, 3, 1) * rik.reshape(-1, 1, 3)).T/ab(rik)**2).T).T/ab(rik))).T

    d12G = lambda rij, rik, i, ij, ik: \
        Dg1(rij, rik).reshape(-1, 3, 1) * Dh2(rik).reshape(-1, 1, 3) + ((hf(rik) * Dg12(rij, rik).T).T)

    Dg12 = lambda rij, rik: \
         (ddg(costh(rij, rik)) * (c1(rij, rik).reshape(-1, 3, 1) * c2(rij, rik).reshape(-1, 1, 3)).T
         + dg(costh(rij, rik)) * dc12(rij, rik).T).T    

    # Helping functions 
    c1 = lambda rij, rik: ((rik.T/ab(rik) - rij.T/ab(rij) * costh(rij, rik)) / ab(rij)).T
    c2 = lambda rij, rik: ((rij.T/ab(rij) - rik.T/ab(rik) * costh(rij, rik)) / ab(rik)).T
    dc11 = lambda rij, rik: \
        ((- c1(rij, rik).reshape(-1, 3, 1) * rij.reshape(-1, 1, 3) \
         - rij.reshape(-1, 3, 1) * c1(rij, rik).reshape(-1, 1, 3) \
         - (costh(rij, rik) * (np.eye(3) -  ((rij.reshape(-1, 1, 3)*rij.reshape(-1, 3, 1)).T/ab(rij)**2).T).T).T \
        ).T/ab(rij)**2).T
    dc22 = lambda rij, rik: \
        ((- c2(rij, rik).reshape(-1, 3, 1) * rik.reshape(-1, 1, 3) \
         - rik.reshape(-1, 3, 1) * c2(rij, rik).reshape(-1, 1, 3) \
         - (costh(rij, rik) * (np.eye(3) -  ((rik.reshape(-1, 1, 3)*rik.reshape(-1, 3, 1)).T/ab(rik)**2).T).T).T \
        ).T/ab(rik)**2).T
    dc12 = lambda rij, rik: \
        (((np.eye(3) -  ((rij.reshape(-1, 1, 3)*rij.reshape(-1, 3, 1)).T/ab(rij)**2).T).T/ab(rij)
         - (c1(rij, rik).reshape(-1, 3, 1) * rik.reshape(-1, 1, 3)).T/ab(rik) \
        )/ab(rik)).T

    costh = lambda rij, rik: np.sum(rij*rik, axis=1) / (ab(rij)*ab(rik))    

    return {
        'atom_type': lambda n: np.zeros_like(n),
        'pair_type': lambda i, j: np.zeros_like(i),
        'F': F,
        'G': G,
        'd1F': d1F,
        'd2F': d2F,
        'd11F': d11F,
        'd12F': d12F,
        'd22F': d22F,
        'd1G': d1G,
        'd2G': d2G,
        'd11G': d11G,
        'd22G': d22G,
        'd12G': d12G,
        'cutoff': a*sigma,
    }



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2021) James Kermode, King's College London
#                       Lars Pastewka, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import itertools

import numpy as np


def mabincount(x, weights, minlength, axis=0):
    """
    Multi-axis bin count. Executes a bin count online a long a specific axis.
    (`numpy.bincount` only works on flattened arrays.)

    Parameters
    ----------
    x : array_like
        Array containing bin indices.
    weights : array_like
        Weights to be binned, dimension `axis` must have same size as x.
    minlength : int
        A minimum number of bins for the output array.
    axis : int, optional
        Axis along which the bin counting is performed. (Default: 0)

    Returns
    -------
    out : np.ndarray
        Array containing the counted data. Array has same dimensions as
        weights, with the exception of dimension `axis` that has is of at least
        `minlength` length.
    """

    # Construct shapes of result array and iterator
    result_shape = list(weights.shape)
    result_shape[axis] = minlength
    iter_shape = list(weights.shape)
    del iter_shape[axis]

    # Initialize result array to zero
    result = np.zeros(result_shape, dtype=weights.dtype)

    # Loop over all trailing dimensions and perform bin count
    for c in itertools.product(*(range(s) for s in iter_shape)):
        axis_slice = list(c)
        axis_slice.insert(axis, slice(None))
        axis_slice = tuple(axis_slice)
        result[axis_slice] = np.bincount(x, weights=weights[axis_slice], minlength=minlength)

    # Return results
    return result



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2017) James Kermode, Warwick University
#                       Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
import numpy as np

from scipy.sparse.linalg import cg

from ase.calculators.calculator import Calculator

from ..elasticity import Voigt_6_to_full_3x3_stress

from ..numpy_tricks import mabincount


class MatscipyCalculator(Calculator):
    def get_hessian(self, atoms, format='sparse', divide_by_masses=False):
        """
        Calculate the Hessian matrix for a pair potential. For an atomic
        configuration with N atoms in d dimensions the hessian matrix is a
        symmetric, hermitian matrix with a shape of (d*N,d*N). The matrix is
        in general a sparse matrix, which consists of dense blocks of shape
        (d,d), which are the mixed second derivatives.

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.
        format: str, optional
            Output format of the Hessian matrix, either 'dense', 'sparse' or
            'neighbour-list'. The format 'sparse' returns a sparse matrix
            representations of scipy. The format 'neighbor-list' returns
            a representation within matscipy's and ASE's neighbor list
            format, i.e. the Hessian is returned per neighbor.
            (Default: 'dense')
        divide_by_masses : bool, optional
            Divided each block entry n the Hessian matrix by sqrt(m_i m_j)
            where m_i and m_j are the masses of the two atoms for the Hessian
            matrix.

        Returns
        -------
        If format=='sparse':
        hessian : scipy.sparse.bsr_matrix
            Hessian matrix in sparse matrix representation
        If format=='neighbor-list'
        hessian_ncc : np.ndarray
            Array containing the Hessian blocks per atom pair
        distances_nc : np.ndarray
            Distance vectors between atom pairs
        """
        raise NotImplementedError

    def get_born_elastic_constants(self, atoms):
        """
        Compute the Born elastic constants. 

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        """
        H_pcc, i_p, j_p, dr_pc, abs_dr_p = self.get_hessian(atoms, 'neighbour-list')

        # Second derivative
        C_pabab = H_pcc.reshape(-1, 3, 1, 3, 1) * dr_pc.reshape(-1, 1, 3, 1, 1) * dr_pc.reshape(-1, 1, 1, 1, 3)
        C_abab = -C_pabab.sum(axis=0) / (2*atoms.get_volume())

        # Symmetrize elastic constant tensor
        C_abab = (C_abab + C_abab.swapaxes(0, 1) + C_abab.swapaxes(2, 3) + C_abab.swapaxes(0, 1).swapaxes(2, 3)) / 4

        return C_abab

    def get_stress_contribution_to_elastic_constants(self, atoms):
        """
        Compute the correction to the elastic constants due to non-zero stress in the configuration.
        Stress term  results from working with the Cauchy stress.

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        """
        stress_cc = Voigt_6_to_full_3x3_stress(self.get_stress())
        delta_cc = np.identity(3)
        C_abab = delta_cc.reshape(3, 1, 3, 1) * stress_cc.reshape(1, 3, 1, 3) - \
                  (delta_cc.reshape(3, 3, 1, 1) * stress_cc.reshape(1, 1, 3, 3) + \
                   delta_cc.reshape(1, 1, 3, 3) * stress_cc.reshape(3, 3, 1, 1)) / 2

        # Symmetrize elastic constant tensor
        C_abab = (C_abab + C_abab.swapaxes(0, 1) + C_abab.swapaxes(2, 3) + C_abab.swapaxes(0, 1).swapaxes(2, 3)) / 4

        return C_abab

    def get_birch_coefficients(self, atoms):
        """
        Compute the Birch coefficients (Effective elastic constants at non-zero stress). 
        
        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        """
        if self.atoms is None:
            self.atoms = atoms

        # Born (affine) elastic constants
        calculator = atoms.get_calculator()
        bornC_abab = calculator.get_born_elastic_constants(atoms)

        # Stress contribution to elastic constants
        stressC_abab = calculator.get_stress_contribution_to_elastic_constants(atoms)

        return bornC_abab + stressC_abab
        

    def get_nonaffine_forces(self, atoms):
        """
        Compute the non-affine forces which result from an affine deformation of atoms.

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        """
        nat = len(atoms)

        H_pcc, i_p, j_p, dr_pc, abs_dr_p = self.get_hessian(atoms, 'neighbour-list')

        naF_pcab = -0.5 * H_pcc.reshape(-1, 3, 3, 1) * dr_pc.reshape(-1, 1, 1, 3)

        naforces_icab = mabincount(i_p, naF_pcab, nat) - mabincount(j_p, naF_pcab, nat)

        return naforces_icab

    def get_non_affine_contribution_to_elastic_constants(self, atoms, eigenvalues=None, eigenvectors=None, tol=1e-5):
        """
        Compute the correction of non-affine displacements to the elasticity tensor.
        The computation of the occuring inverse of the Hessian matrix is bypassed by using a cg solver.

        If eigenvalues and and eigenvectors are given the inverse of the Hessian can be easily computed.


        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        eigenvalues: array
            Eigenvalues in ascending order obtained by diagonalization of Hessian matrix.
            If given 

        eigenvectors: array
            Eigenvectors corresponding to eigenvalues.

        tol: float
            Tolerance for the conjugate-gradient solver. 

        """

        nat = len(atoms)

        calc = atoms.get_calculator()

        C_abab = np.zeros((3,3,3,3))        

        if (eigenvalues is not None) and (eigenvectors is not None):
            naforces_icab = calc.get_nonaffine_forces(atoms)

            G_incc = (eigenvectors.T).reshape(-1, 3*nat, 1, 1) * naforces_icab.reshape(1, 3*nat, 3, 3)
            G_incc = (G_incc.T/np.sqrt(eigenvalues)).T
            G_icc  = np.sum(G_incc, axis=1)
            C_abab = np.sum(G_icc.reshape(-1,3,3,1,1) * G_icc.reshape(-1,1,1,3,3), axis=0)

        else:
            H_nn = calc.get_hessian(atoms, "sparse")
            naforces_icab = calc.get_nonaffine_forces(atoms)

            D_iab = np.zeros((3*nat, 3, 3))
            for i in range(3):
                for j in range(3):
                    x, info = cg(H_nn, naforces_icab[:, :, i, j].flatten(), atol=tol)
                    if info != 0:
                        raise RuntimeError(" info > 0: CG tolerance not achieved, info < 0: Exceeded number of iterations.")
                    D_iab[:,i,j] = x

            C_abab = np.sum(naforces_icab.reshape(3*nat, 3, 3, 1, 1) * D_iab.reshape(3*nat, 1, 1, 3, 3), axis=0)
        
        # Symmetrize 
        C_abab = (C_abab + C_abab.swapaxes(0, 1) + C_abab.swapaxes(2, 3) + C_abab.swapaxes(0, 1).swapaxes(2, 3)) / 4             

        return -C_abab/atoms.get_volume()

    def get_numerical_non_affine_forces(self, atoms, d=1e-6):
        """

        Calculate numerical non-affine forces using central finite differences.
        This is done by deforming the box, rescaling atoms and measure the force.

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        """

        nat = len(atoms)
        cell = atoms.cell.copy()
        fna_ncc = np.zeros((nat, 3, 3, 3))

        for i in range(3):
            # Diagonal 
            x = np.eye(3)
            x[i, i] += d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            fplus = atoms.get_forces()

            x[i, i] -= 2 * d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            fminus = atoms.get_forces()

            naForces_ncc = (fplus - fminus) / (2 * d)
            fna_ncc[:, 0, i, i] = naForces_ncc[:, 0]
            fna_ncc[:, 1, i, i] = naForces_ncc[:, 1]
            fna_ncc[:, 2, i, i] = naForces_ncc[:, 2]

            # Off diagonal
            j = i - 2
            x[i, j] = d
            x[j, i] = d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            fplus = atoms.get_forces()

            x[i, j] = -d
            x[j, i] = -d
            atoms.set_cell(np.dot(cell, x), scale_atoms=True)
            fminus = atoms.get_forces()

            naForces_ncc = (fplus - fminus) / (4 * d)
            fna_ncc[:, 0, i, j] = naForces_ncc[:, 0]
            fna_ncc[:, 0, j, i] = naForces_ncc[:, 0]
            fna_ncc[:, 1, i, j] = naForces_ncc[:, 1]
            fna_ncc[:, 1, j, i] = naForces_ncc[:, 1]
            fna_ncc[:, 2, i, j] = naForces_ncc[:, 2]
            fna_ncc[:, 2, j, i] = naForces_ncc[:, 2]

        return fna_ncc



#!/usr/bin/env python3
# PYTHON_ARGCOMPLETE_OK
# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2019) Johannes Hoermann, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================
"""Solves 1D Poisson-Nernst-Planck system

Copyright 2019 IMTEK Simulation
University of Freiburg

Authors:
  Johannes Hoermann <johannes.hoermann@imtek-uni-freiburg.de>
"""
import argparse
import datetime
import logging
import os
import sys
import numpy as np

def main():
    """Solve Poisson-Nernst-Planck system and store distribution.

    Specify quantities in SI units at this command line interface."""


    # in order to have both:
    # * preformatted help text and ...
    # * automatic display of defaults
    class ArgumentDefaultsAndRawDescriptionHelpFormatter(
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.RawDescriptionHelpFormatter):
        pass

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=ArgumentDefaultsAndRawDescriptionHelpFormatter)

    parser.add_argument('outfile', metavar='OUT', default=None, nargs='?',
                        help='binary numpy .npz or plain text .txt output file')

    # physical system parameters
    parser.add_argument('--concentrations', '-c',
                        default=[0.1, 0.1], type=float, nargs='+',
                        metavar='c', required=False, dest="concentrations",
                        help='Ion species concentrations c (mol m^-3, mM)')

    parser.add_argument('--charges', '-z',
                        default=[1, -1], type=float, nargs='+',
                        metavar='z', required=False, dest="charges",
                        help='Ion species number charges z')

    parser.add_argument('--potential', '-u',
                        default=0.05, type=float,
                        metavar='U', required=False, dest="potential",
                        help='Potential drop from left to right dU (V)')

    parser.add_argument('--length', '-l',
                        default=100.0e-9, type=float,
                        metavar='L', required=False, dest="length",
                        help='Domain length (m)')

    parser.add_argument('--temperature', '-T',
                        default=298.15, type=float,
                        metavar='T', required=False, dest="temperature",
                        help='Temperature (K)')

    parser.add_argument('--relative-permittivity', '--epsilon-r', '--eps',
                        default=79.0, type=float,
                        metavar='eps', required=False,
                        dest="relative_permittivity",
                        help='Relative permittivity')

    parser.add_argument('--compact-layer', '--stern-layer', '--lambda-s',
                        default=0.0, type=float,
                        metavar='L', required=False,
                        dest="lambda_S",
                        help='Stern or compact layer thickness (for Robin BC)')

    parser.add_argument('--boundary-conditions', '-bc',
                        default='cell', type=str,
                        metavar='BC', required=False,
                        dest="boundary_conditions",
                        choices=(
                            'interface',  # open half-space
                            'cell',  # 1D electorchemical cell with zero flux BC
                            'cell-stern',  # 1D cell with linear compact layer regime
                            'cell-robin',  # 1D cell with implict compact layer by Robin BC
                            ),
                        help='Boundary conditions')

    # technical settings
    parser.add_argument('--segments', '-N',
                        default=200, type=int,
                        metavar='N', required=False,
                        dest="segments",
                        help='Number of discretization segments')

    parser.add_argument('--maximum-iterations', '--maxit',
                        default=20, type=int,
                        metavar='N', required=False,
                        dest="maxit",
                        help='Maximum number of Newton iterations')

    parser.add_argument('--absolute-tolerance',
                        default=1e-8, type=float,
                        metavar='e', required=False,
                        dest="absolute_tolerance",
                        help='Absolute tolerance Newton solver convergence criterion')

    parser.add_argument('--convergence-stats', default=False, required=False,
                        action='store_true', dest="convergence",
                        help='Record and store Newton solver convergence statistics')

    parser.add_argument('--debug', default=False, required=False,
                        action='store_true', dest="debug", help='debug flag')
    parser.add_argument('--verbose', default=False, required=False,
                        action='store_true', dest="verbose", help='verbose flag')
    parser.add_argument('--log', required=False, nargs='?', dest="log",
                        default=None, const='pnp.log', metavar='LOG',
                        help='Write log file pnp.log, optionally specify log file name')

    try:
        import argcomplete
        argcomplete.autocomplete(parser)
        # This supports bash autocompletion. To enable this, pip install
        # argcomplete, activate global completion, or add
        #      eval "$(register-python-argcomplete lpad)"
        # into your .bash_profile or .bashrc
    except ImportError:
        pass

    args = parser.parse_args()

    if args.debug:
        loglevel = logging.DEBUG
    elif args.verbose:
        loglevel = logging.INFO
    else:
        loglevel = logging.WARNING

    # PoissonNernstPlanckSystem makes extensive use of Python's logging module
    logformat = "[ %(filename)s:%(lineno)s - %(funcName)s() ]: %(message)s"

    logging.basicConfig(level=loglevel,
                        format=logformat)

    # explicitly modify the root logger (necessary?)
    logger = logging.getLogger()
    logger.setLevel(loglevel)

    # remove all handlers
    for h in logger.handlers:
        logger.removeHandler(h)

    # create and append custom handles
    ch = logging.StreamHandler()
    formatter = logging.Formatter(logformat)
    ch.setFormatter(formatter)
    ch.setLevel(loglevel)
    logger.addHandler(ch)

    if args.log:
        fh = logging.FileHandler(args.log)
        fh.setFormatter(formatter)
        fh.setLevel(loglevel)
        logger.addHandler(fh)

    # use FEniCS finite element solver if available,
    # otherwise own controlled volume scheme
    try:
        import fenics
        from matscipy.electrochemistry.poisson_nernst_planck_solver_fenics \
            import PoissonNernstPlanckSystemFEniCS as PoissonNernstPlanckSystem
        import dolfin
        dolfin.cpp.log.set_log_level(loglevel)
        logger.info("Will use FEniCS finite element solver.")
    except ModuleNotFoundError:
        logger.warning(
            "No FEniCS finite element solver found,"
            " falling back to internal controlled-volume implementation."
            " ATTENTION: Number conservation not exact.")
        from matscipy.electrochemistry import PoissonNernstPlanckSystem

    # set up system
    pnp = PoissonNernstPlanckSystem(
        c=np.array(args.concentrations, dtype=float),
        z=np.array(args.charges, dtype=float),
        L=float(args.length),
        T=float(args.temperature),
        delta_u=float(args.potential),
        lambda_S=float(args.lambda_S),
        N=args.segments,
        maxit=args.maxit,
        e=args.absolute_tolerance,
        relative_permittivity=float(args.relative_permittivity))

    if args.boundary_conditions in ('cell-robin') and float(args.lambda_S) > 0:
        pnp.useSternLayerCellBC()
    elif ((args.boundary_conditions in ('cell-robin') and float(args.lambda_S) == 0)
            or args.boundary_conditions == 'cell'):
        pnp.useStandardCellBC()
    elif args.boundary_conditions == 'interface':
        pnp.useStandardInterfaceBC()
    else:
        raise ValueError("Boundary conditions '{}' not implemented!".format(
                         args.boundary_conditions))

    pnp.solve()

    extra_kwargs = {}
    if args.convergence:
        extra_kwargs.update({
            'convergence_step_absolute': pnp.convergenceStepAbsolute,
            'convergence_step_relative': pnp.convergenceStepRelative,
            'convergence_residual_absolute': pnp.convergenceResidualAbsolute})

    if not args.outfile:
        outfile = sys.stdout
        format = 'txt'
    else:
        outfile = args.outfile
        _, format = os.path.splitext(outfile)

    if format == '.npz':
        np.savez(file=outfile,
                 x=pnp.grid, u=pnp.potential, c=pnp.concentration, **extra_kwargs)
    else:  # elif format == '.txt'
        comment = '\n'.join((
            'Poisson-Nernst-Planck system, generated on {:s}, {:s}'.format(
                os.uname().nodename,
                datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')),
            'All quantities in SI units.',
            'grid x (m), potential u (V), concentration c (mM)'))
        header = comment + '\n' + '{:20s} {:22s} '.format('x', 'u') + ' '.join(
            ['{:22s}'.format('c{:02d}'.format(k)) for k in range(pnp.M)])
        data = np.column_stack([pnp.grid, pnp.potential, pnp.concentration.T])
        np.savetxt(outfile, data, fmt='%22.15e', header=header)

    # write out final state as usual, but mark process failed if not converged
    # if not pnp.converged:
    #    sys.exit(1)

    sys.exit(0)


if __name__ == '__main__':
    main()



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014-2018) James Kermode, King's College London
#                       Lars Pastewka, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import numpy as np

from scipy.special import factorial2
from scipy.sparse import bsr_matrix

import ase

from ...neighbours import neighbour_list, first_neighbours
from ..calculator import MatscipyCalculator
from ...numpy_tricks import mabincount

###


class InversePowerLawPotential():
    """
    Functional form for a smoothed inverse-power-law potential (IPL)
    with an repulsive exponent of 10.

    Parameters
    ----------
    epsilon : float
        Energy scale
    cutoff : float
        Cutoff for the pair-interaction
    minSize : float
        Minimal size of a particle, lower bound of distribtuion
    maxSize : float
        Maximal size of a particle, upper bound of distribtuion
    na : float
        Non-additivity paramter for pairwise sizes
    q : int
        Smooth the potential up to the q-th derivative.
        For q=0 the potential is smoothed, for q=1 the potential
        and its first derivative are zero at the cutoff,...

    Reference:
    ----------
        E. Lerner, Journal of Non-Crystalline Solids, 522, 119570.
    """

    def __init__(self, epsilon, cutoff, na, q, minSize, maxSize):
        self.epsilon = epsilon
        self.cutoff = cutoff
        self.minSize = minSize
        self.maxSize = maxSize
        self.na = na
        self.q = q
        self.coeffs = []
        for index in range(0, q+1):
            first_expr = np.power(-1, index+1) / (factorial2(
                2*q - 2*index, exact=True) * factorial2(2*index, exact=True))
            second_expr = factorial2(10+2*q, exact=True) / (factorial2(
                8, exact=True) * (10+2*index))
            third_expr = np.power(cutoff, -(10+2*index))
            self.coeffs.append(first_expr * second_expr * third_expr)

    def __call__(self, r, ijsize):
        """
        Return function value (potential energy)
        """
        ipl = self.epsilon * np.power(ijsize, 10) / np.power(r, 10)
        for l in range(0, self.q+1):
            ipl += self.epsilon * self.coeffs[l] * np.power(r/ijsize, 2*l)

        return ipl

    def mix_sizes(self, isize, jsize):
        """
        Nonadditive interaction rule for the cross size of particles i and j.
        """
        return 0.5 * (isize+jsize) * (1 - self.na * np.absolute(isize-jsize))

    def get_cutoff(self):
        """
        Return the cutoff.
        """
        return self.cutoff

    def get_coeffs(self):
        """
        Return the smoothing coefficients of the potential.
        """
        return self.coeffs

    def get_maxSize(self):
        """
        Return the maximal size of a particle (=Upper boundary of distribution)
        """
        return self.maxSize

    def get_minSize(self):
        """
        Return the minimal size of a particle (=Lower boundary of distribution)
        """
        return self.minSize

    def first_derivative(self, r, ijsize):
        """
        Return first derivative
        """
        dipl = -10 * self.epsilon * np.power(ijsize, 10) / np.power(r, 11)
        for l in range(0, self.q+1):
            dipl += 2*self.epsilon*l * \
                self.coeffs[l] * np.power(r, 2*l-1) / np.power(ijsize, 2*l)

        return dipl

    def second_derivative(self, r, ijsize):
        """
        Return second derivative
        """
        ddipl = 110 * self.epsilon * np.power(ijsize, 10) / np.power(r, 12)
        for l in range(0, self.q+1):
            ddipl += self.epsilon * \
                (4*np.power(l, 2)-2*l) * \
                self.coeffs[l] * np.power(r, 2*l-2) / np.power(ijsize, 2*l)

        return ddipl

    def derivative(self, n=1):
        if n == 1:
            return self.first_derivative
        elif n == 2:
            return self.second_derivative
        else:
            raise ValueError(
                "Don't know how to compute {}-th derivative.".format(n))

###


class Polydisperse(MatscipyCalculator):
    implemented_properties = ["energy", "free_energy", "stress", "forces", "hessian"]
    default_parameters = {}
    name = "Polydisperse"

    def __init__(self, f, cutoff=None):
        MatscipyCalculator.__init__(self)
        self.f = f

    def calculate(self, atoms, properties, system_changes):
        MatscipyCalculator.calculate(self, atoms, properties, system_changes)

        f = self.f
        nb_atoms = len(self.atoms)
        if atoms.has("size"):
            size = self.atoms.get_array("size")
        else:
            raise AttributeError(
                "Attribute error: Unable to load atom sizes from atoms object!")

        i_p, j_p, r_pc, r_p = neighbour_list("ijDd", self.atoms, f.get_maxSize()*f.get_cutoff())
        ijsize = f.mix_sizes(size[i_p], size[j_p])

        # Mask neighbour list to consider only true neighbors
        mask = r_p <= f.get_cutoff() * ijsize
        i_p = i_p[mask]
        j_p = j_p[mask]
        r_pc = r_pc[mask]
        r_p = r_p[mask]
        ijsize = ijsize[mask]
        e_p = f(r_p, ijsize)
        de_p = f.first_derivative(r_p, ijsize)

        # Energy
        epot = 0.5*np.sum(e_p)

        # Forces
        df_pc = -0.5*de_p.reshape(-1, 1)*r_pc/r_p.reshape(-1, 1)

        f_nc = mabincount(j_p, df_pc, nb_atoms) - mabincount(i_p, df_pc, nb_atoms)

        # Virial
        virial_v = -np.array([r_pc[:, 0] * df_pc[:, 0],               # xx
                              r_pc[:, 1] * df_pc[:, 1],               # yy
                              r_pc[:, 2] * df_pc[:, 2],               # zz
                              r_pc[:, 1] * df_pc[:, 2],               # yz
                              r_pc[:, 0] * df_pc[:, 2],               # xz
                              r_pc[:, 0] * df_pc[:, 1]]).sum(axis=1)  # xy

        self.results = {'energy': epot,
                        'free_energy': epot,
                        'stress': virial_v/self.atoms.get_volume(),
                        'forces': f_nc}

    ###

    def get_hessian(self, atoms, format='sparse', divide_by_masses=False):
        """
        Calculate the Hessian matrix for a polydisperse systems where atoms interact via a pair potential.
        For an atomic configuration with N atoms in d dimensions the hessian matrix is a symmetric, hermitian matrix
        with a shape of (d*N,d*N). The matrix is due to the cutoff function a sparse matrix, which consists of dense blocks of shape (d,d), which
        are the mixed second derivatives. The result of the derivation for a pair potential can be found in:
        L. Pastewka et. al. "Seamless elastic boundaries for atomistic calculations", Phys. Ev. B 86, 075459 (2012).

        Parameters
        ----------
        atoms: ase.Atoms
            Atomic configuration in a local or global minima.

        format: "sparse" or "neighbour-list"
            Output format of the hessian matrix.

        divide_by_masses: bool
            Divide the block "l,m" by the corresponding atomic masses "sqrt(m_l, m_m)" to obtain dynamical matrix.

        Restrictions
        ----------
        This method is currently only implemented for three dimensional systems

        """

        if self.atoms is None:
            self.atoms = atoms

        f = self.f
        nb_atoms = len(self.atoms)
        if atoms.has("size"):
            size = self.atoms.get_array("size")
        else:
            raise AttributeError(
                "Attribute error: Unable to load atom sizes from atoms object! Probably missing size array.")

        i_p, j_p, r_pc, r_p = neighbour_list("ijDd", self.atoms, f.get_maxSize()*f.get_cutoff())
        ijsize = f.mix_sizes(size[i_p], size[j_p])

        # Mask neighbour list to consider only true neighbors
        mask = r_p <= f.get_cutoff()*ijsize
        i_p = i_p[mask]
        j_p = j_p[mask]
        r_pc = r_pc[mask]
        r_p = r_p[mask]
        ijsize = ijsize[mask]
        first_i = first_neighbours(nb_atoms, i_p)

        if divide_by_masses:
            mass_n = self.atoms.get_masses()
            geom_mean_mass_p = np.sqrt(mass_n[i_p]*mass_n[j_p])

        # Hessian 
        de_p = f.first_derivative(r_p, ijsize)
        dde_p = f.second_derivative(r_p, ijsize)
        n_pc = (r_pc.T/r_p).T
        H_pcc = -(dde_p * (n_pc.reshape(-1, 3, 1)
                           * n_pc.reshape(-1, 1, 3)).T).T
        H_pcc += -(de_p/r_p * (np.eye(3, dtype=n_pc.dtype)
                                   - (n_pc.reshape(-1, 3, 1) * n_pc.reshape(-1, 1, 3))).T).T

        if format == "sparse":
            if divide_by_masses:
                H = bsr_matrix(((H_pcc.T/geom_mean_mass_p).T,
                                j_p, first_i), shape=(3*nb_atoms, 3*nb_atoms))

            else:
                H = bsr_matrix((H_pcc, j_p, first_i), shape=(3*nb_atoms, 3*nb_atoms))

            Hdiag_icc = np.empty((nb_atoms, 3, 3))
            for x in range(3):
                for y in range(3):
                    Hdiag_icc[:, x, y] = - \
                        np.bincount(i_p, weights=H_pcc[:, x, y])

            if divide_by_masses:
                H += bsr_matrix(((Hdiag_icc.T/mass_n).T, np.arange(nb_atoms),
                        np.arange(nb_atoms+1)), shape=(3*nb_atoms, 3*nb_atoms))         

            else:
                H += bsr_matrix((Hdiag_icc, np.arange(nb_atoms),
                         np.arange(nb_atoms+1)), shape=(3*nb_atoms, 3*nb_atoms))

            return H

        # Neighbour list format
        elif format == "neighbour-list":
            return H_pcc, i_p, j_p, r_pc, r_p



# -*- coding: utf-8 -*-
# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2020) Wolfram Nöhring, University of Freiburg
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

from .io import EAMParameters
import numpy as np


def average_potential(
    concentrations,
    parameters,
    F,
    f,
    rep,
    kind="eam/alloy",
    avg_atom="A",
    atomic_number=999,
    crystal_structure="unknown",
    lattice_constant=1.0,
):
    r"""Generate Average-atom potential

    The Average-atom (A-atom) potential is a mean-field approximation
    for random alloys, see Ref. `1`_. The purpose is to replace the true
    elements by a single fictious element, the A-atom. A configuration
    of A-atoms yields approximately the same potential energy as the
    corresponding random alloy configuration. Other average material
    properties, e.g. the elastic constants, are reproduced as well.
    For a full derivation, see Ref. `1`. 

    The A-atom potential has standard EAM form, i.e. it can be tabulated
    just like any other EAM potential. The potential functions are simply
    the concentration-weighted averages of the pure element functions:

    .. math::
        
        \phi_{AA}\left(r_{\gamma\delta}\right) 
            &= \sum_{X}^{N_T}\sum_{Y}^{N_T} c_{X}c_{Y}  
            \phi_{XY}\left(r_{\gamma\delta}\right) \quad\text{(pair potential)}, \\
        U_{A}\left(\rho_\gamma\right)  
            &= \sum_{X}^{N_T}c_{X}U_{X}\left(\rho_\gamma\right) \quad\text{(embedding energy)}, \\
        g_A\left(r_{\gamma\delta}\right) 
            &= \sum_{X}^{N_T}c_X g_X\left(r_{\gamma\delta}\right)\quad\text{(electron density)},\;\text{and}\\
        m_A &= \sum_{X}^{N_T}c_X m_X\quad\text{(mass)}.

    .. note::
        
        Currently, only eam/alloy-style potentials can be averaged.
        The extension to eam/fs should be straightforward, however.


    Parameters
    ----------
    concentrations: array_like
        concentrations of the elements in the A-atom
    parameters: EAMParameters
        EAM potential parameters
    F : array_like
        tabulated embedding energy functionals
    f : array_like
        tabulated electron density functions
    rep : array_like
        tabulated pair potential energy functions

    Returns
    -------
    parameters : EAMParameters
        EAM potential parameters
    new_F : array_like
        tabulated embedding energy functionals, including A-atom functional
    new_f : array_like
        tabulated electron density functions, including A-atom function(s)
    new_rep : array_like
        tabulated pair potential energy functions, including pairs with A-atom

    Examples
    --------
    >>> from matscipy.calculators.eam import io, average_atom
    >>> source, parameters, F, f, rep = io.read_eam(
    >>>     "ZrCu.onecolumn.eam.alloy"
    >>> )
    >>> concentrations = [0.5, 0.5]
    >>> (new_parameters, new_F, new_f, new_rep) = average_atom.average_potential(
    >>>     concentrations, parameters, F, f, rep
    >>> )
    >>> composition = " ".join(
    >>>     [str(c * 100.0) + "% {0},".format(e) for c, e in zip(concentrations, parameters.symbols)]
    >>> )
    >>> composition = composition.rstrip(",")
    >>> source += ", averaged for composition {0}".format(composition)
    >>> io.write_eam(
    >>>     source,
    >>>     new_parameters,
    >>>     new_F,
    >>>     new_f,
    >>>     new_rep,
    >>>     "ZrCu.onecolumn.averaged.eam.alloy",
    >>>     kind="eam/alloy",
    >>> )

    Read an EAM potential table for two elements in eam/alloy format, and create
    a new table with additional A-atom functions for the equicomposition alloy.

    References
    ----------

    .. [1] Varvenne, C., Luque, A., Nöhring, W. G. & Curtin, W. A. 
       Average-atom interatomic potential for random alloys. 
       Physical Review B 93, 104201 (2016).

    Notes
    -----
    Notation:
     * :math:`N` Number of atoms 
     * :math:`N_T` Number of elements
     * :math:`r_{\nu\mu{}}` Pair distance of atoms :math:`\nu` and :math:`\mu`
     * :math:`\phi_{\nu\mu}(r_{\nu\mu{}})` Pair potential energy of atoms :math:`\nu` and :math:`\mu` 
     * :math:`\rho_{\nu}` Total electron density of atom :math:`\nu`  
     * :math:`U_\nu(\rho_nu)` Embedding energy of atom :math:`\nu` 
     * :math:`g_{\delta}\left(r_{\gamma\delta}\right) \equiv g_{\gamma\delta}` Contribution from atom :math:`\delta` to :math:`\rho_\gamma`
     * :math:`m_\nu` mass of atom :math:`\nu`

    """

    if kind == "eam" or kind == "eam/fs":
        raise NotImplementedError
    assert np.isclose(np.sum(concentrations), 1)
    symbols = [s for s in parameters.symbols] + [avg_atom]
    atomic_numbers = np.r_[parameters.atomic_numbers, atomic_number]
    atomic_masses = np.r_[
        parameters.atomic_masses, np.average(parameters[2], weights=concentrations)
    ]
    lattice_constants = np.r_[parameters.lattice_constants, lattice_constant]
    crystal_structures = np.r_[
        parameters.crystal_structures, np.array(crystal_structure)
    ]
    new_parameters = EAMParameters(
        symbols,
        atomic_numbers,
        atomic_masses,
        lattice_constants,
        crystal_structures,
        parameters.number_of_density_grid_points,
        parameters.number_of_distance_grid_points,
        parameters.density_grid_spacing,
        parameters.distance_grid_spacing,
        parameters.cutoff,
    )
    # Average embedding energy and electron density functions
    new_F = np.r_[F, np.zeros((1, F.shape[1]), dtype=F.dtype)]
    new_f = np.r_[f, np.zeros((1, f.shape[1]), dtype=f.dtype)]
    new_F[-1, :] = np.average(F, axis=0, weights=concentrations)
    new_f[-1, :] = np.average(f, axis=0, weights=concentrations)
    # Average the pair potential
    new_rep = np.concatenate(
        (rep, np.zeros((rep.shape[0], 1, rep.shape[2]), dtype=rep.dtype)), axis=1
    )
    new_rep = np.concatenate(
        (new_rep, np.zeros((1, new_rep.shape[1], rep.shape[2]), dtype=rep.dtype)),
        axis=0,
    )
    # Consider the matrix Vij of pair functions
    # i: rows, each corresponding to an element
    # j: columns, each corresponding to an element
    # Each element corresponds to the function
    # for the atom pair i,j
    #
    # Compute the last row of the new Vij matrix, excluding the
    # value on the diagonal. Each column j corresponds to the
    # interaction of a particular type j with the homogenous material.
    new_rep[-1, :-1, :] = np.average(rep, axis=0, weights=concentrations)
    new_rep[:-1, -1, :] = new_rep[-1, :-1, :]
    # Compute the last potential on the diagonal. This is the
    # interaction of the homogeneous material with itself.
    column = new_rep[:-1, -1, :]
    new_rep[-1, -1, :] = np.average(column, axis=0, weights=concentrations)

    return new_parameters, new_F, new_f, new_rep



# ======================================================================
# matscipy - Python materials science tools
# https://github.com/libAtoms/matscipy
#
# Copyright (2014) James Kermode, King's College London
#                  Lars Pastewka, Karlsruhe Institute of Technology
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
# ======================================================================

import numpy as np

from ase.data import atomic_numbers

import _matscipy
from _matscipy import first_neighbours, get_jump_indicies

###


def mic(dr, cell, pbc=None):
    """
    Apply minimum image convention to an array of distance vectors.

    Parameters
    ----------
    dr : array_like
        Array of distance vectors.
    cell : array_like
        Simulation cell.
    pbc : array_like, optional
        Periodic boundary conditions in x-, y- and z-direction. Default is to
        assume periodic boundaries in all directions.

    Returns
    -------
    dr : array
        Array of distance vectors, wrapped according to the minimum image
        convention.
    """
    # Check where distance larger than 1/2 cell. Particles have crossed
    # periodic boundaries then and need to be unwrapped.
    rec = np.linalg.inv(cell)
    if pbc is not None:
        rec *= np.array(pbc, dtype=int).reshape(3, 1)
    dri = np.round(np.dot(dr, rec))

    # Unwrap
    return dr - np.dot(dri, cell)


def neighbour_list(quantities, atoms=None, cutoff=None, positions=None,
                   cell=None, pbc=None, numbers=None, cell_origin=None):
    """
    Compute a neighbor list for an atomic configuration. Atoms outside periodic
    boundaries are mapped into the box. Atoms outside nonperiodic boundaries
    are included in the neighbor list but the complexity of neighbor list search
    for those can become n^2.

    The neighbor list is sorted by first atom index 'i', but not by second
    atom index 'j'.

    The neighbour list accepts either an ASE Atoms object or positions and cell
    vectors individually.

    Parameters
    ----------
    quantities : str
        Quantities to compute by the neighbor list algorithm. Each character
        in this string defines a quantity. They are returned in a tuple of
        the same order. Possible quantities are
            'i' : first atom index
            'j' : second atom index
            'd' : absolute distance
            'D' : distance vector
            'S' : shift vector (number of cell boundaries crossed by the bond
                  between atom i and j). With the shift vector S, the
                  distances D between atoms can be computed from:
                  D = a.positions[j]-a.positions[i]+S.dot(a.cell)
    atoms : ase.Atoms
        Atomic configuration. (Default: None)
    cutoff : float or dict
        Cutoff for neighbor search. It can be
            - A single float: This is a global cutoff for all elements.
            - A dictionary: This specifies cutoff values for element
              pairs. Specification accepts element numbers of symbols.
              Example: {(1, 6): 1.1, (1, 1): 1.0, ('C', 'C'): 1.85}
            - A list/array with a per atom value: This specifies the radius of
              an atomic sphere for each atoms. If spheres overlap, atoms are
              within each others neighborhood.
    positions : array_like
        Atomic positions. (Default: None)
    cell : array_like
        Cell vectors as a 3x3 matrix. (Default: Shrink wrapped cell)
    pbc : array_like
        3-vector containing periodic boundary conditions in all three
        directions. (Default: Nonperiodic box)
    numbers : array_like
        Array containing the atomic numbers.

    Returns
    -------
    i, j, ... : array
        Tuple with arrays for each quantity specified above. Indices in `i`
        are returned in ascending order 0..len(a), but the order of (i,j)
        pairs is not guaranteed.

    Examples
    --------
    Examples assume Atoms object *a* and numpy imported as *np*.
    1. Coordination counting:
        i = neighbor_list('i', a, 1.85)
        coord = np.bincount(i)

    2. Coordination counting with different cutoffs for each pair of species
        i = neighbor_list('i', a,
                          {('H', 'H'): 1.1, ('C', 'H'): 1.3, ('C', 'C'): 1.85})
        coord = np.bincount(i)

    3. Pair distribution function:
        d = neighbor_list('d', a, 10.00)
        h, bin_edges = np.histogram(d, bins=100)
        pdf = h/(4*np.pi/3*(bin_edges[1:]**3 - bin_edges[:-1]**3)) * a.get_volume()/len(a)

    4. Pair potential:
        i, j, d, D = neighbor_list('ijdD', a, 5.0)
        energy = (-C/d**6).sum()
        pair_forces = (6*C/d**5  * (D/d).T).T
        forces_x = np.bincount(j, weights=pair_forces[:, 0], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 0], minlength=len(a))
        forces_y = np.bincount(j, weights=pair_forces[:, 1], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 1], minlength=len(a))
        forces_z = np.bincount(j, weights=pair_forces[:, 2], minlength=len(a)) - \
                   np.bincount(i, weights=pair_forces[:, 2], minlength=len(a))

    5. Dynamical matrix for a pair potential stored in a block sparse format:
        from scipy.sparse import bsr_matrix
        i, j, dr, abs_dr = neighbor_list('ijDd', atoms)
        energy = (dr.T / abs_dr).T
        dynmat = -(dde * (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3)).T).T \
                 -(de / abs_dr * (np.eye(3, dtype=energy.dtype) - \
                   (energy.reshape(-1, 3, 1) * energy.reshape(-1, 1, 3))).T).T
        dynmat_bsr = bsr_matrix((dynmat, j, first_i), shape=(3*len(a), 3*len(a)))

        dynmat_diag = np.empty((len(a), 3, 3))
        for x in range(3):
            for y in range(3):
                dynmat_diag[:, x, y] = -np.bincount(i, weights=dynmat[:, x, y])

        dynmat_bsr += bsr_matrix((dynmat_diag, np.arange(len(a)),
                                  np.arange(len(a) + 1)),
                                 shape=(3 * len(a), 3 * len(a)))


  i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms, dict)
e_nc = (dr_nc.T/abs_dr_n).T
    D_ncc = -(dde_n * (e_nc.reshape(-1,3,1) * e_nc.reshape(-1,1,3)).T).T
    D_ncc += -(de_n/abs_dr_n * (np.eye(3, dtype=e_nc.dtype) - (e_nc.reshape(-1,3,1) * e_nc.reshape(-1,1,3))).T).T

    D = bsr_matrix((D_ncc, j_n, first_i), shape=(3*nat,3*nat))

    Ddiag_icc = np.empty((nat,3,3))
    for x in range(3):
        for y in range(3):
            Ddiag_icc[:,x,y] = -np.bincount(i_n, weights = D_ncc[:,x,y])

    D += bsr_matrix((Ddiag_icc,np.arange(nat),np.arange(nat+1)), shape=(3*nat,3*nat))

    return D
    """

    if cutoff is None:
        raise ValueError('Please provide a value for the cutoff radius.')

    if atoms is None:
        if positions is None:
            raise ValueError('You provided neither an ASE Atoms object nor '
                             'a positions array.')
        if cell is None:
            # Shrink wrapped cell
            rmin = np.min(positions, axis=0)
            rmax = np.max(positions, axis=0)
            cell_origin = rmin
            cell = np.diag(rmax - rmin)
        if cell_origin is None:
            cell_origin = np.zeros(3)
        if pbc is None:
            pbc = np.zeros(3, dtype=bool)
        if numbers is None:
            numbers = np.ones(len(positions), dtype=np.int32)
    else:
        if positions is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'individual position atomic positions at the same '
                             'time.')
        positions = atoms.positions
        if cell_origin is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'a cell origin at the same time.')
        cell_origin = np.zeros(3)
        if cell is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'cell vectors at the same time.')
        cell = atoms.cell
        if pbc is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'separate periodicity information at the same '
                             'time.')
        pbc = atoms.pbc
        if numbers is not None:
            raise ValueError('You cannot provide an ASE Atoms object and '
                             'separate atomic numbers at the same time.')
        numbers = atoms.numbers.astype(np.int32)

    if isinstance(cutoff, dict):
        maxel = np.max(numbers)
        _cutoff = np.zeros([maxel+1, maxel+1], dtype=float)
        for (el1, el2), c in cutoff.items():
            try:
                el1 = atomic_numbers[el1]
            except:
                pass
            try:
                el2 = atomic_numbers[el2]
            except:
                pass
            if el1 < maxel+1 and el2 < maxel+1:
                _cutoff[el1, el2] = c
                _cutoff[el2, el1] = c
    else:
        _cutoff = cutoff

    return _matscipy.neighbour_list(quantities, cell_origin, cell,
                                    np.linalg.inv(cell.T), pbc, positions,
                                    _cutoff, numbers)


def triplet_list(first_neighbours, abs_dr_p=None, cutoff=None, i_p=None, j_p=None):
    """
    Compute a triplet list for an atomic configuration. The triple list is a
    mask that can be applied to the corresponding neighbour list to mask
    triplet properties.
    The triplet list accepts an first_neighbour array (generated by
    first_neighbours) as input.

    Parameters
    ----------
    first_neighbours : array
        adresses of the first time an atom occours in the neighour list

    Returns
    -------
    ij_t, ik_t : array
        lists of adresses that form triples in the pair lists
    jk_t : array (if and only if i_p, j_p, first_i != None)
        list of pairs jk that connect each triplet ij, ik
        between atom j and k

    Example
    -------
    i_n, j_n, abs_dr_p = neighbour_list('ijd', atoms=atoms, cutoff=cutoff)

    first_i = np.array([0, 2, 6, 10], dtype='int32')
    a = triplet_list(first_i, [2.2]*9+[3.0], 2.6)

    # one may obtain first_ij by using
    find_ij = first_neighbours(len(i_p), ij_t)
    # or (slower but less parameters and more general,
    # i.e for every ordered list)
    first_ij = get_jump_indicies(ij_t)

    """
    if not (abs_dr_p is None or cutoff is None):
        res = _matscipy.triplet_list(first_neighbours, abs_dr_p, cutoff)
    else:
        res = _matscipy.triplet_list(first_neighbours)
    # TODO: should be wrapped in c and be independet of i_n
    # and j_n as of j_n is sorted; related issue #50
    # add some tests!!!
    if not (i_p is None or j_p is None or first_neighbours is None):
        ij_t, ik_t = res
        jk_t = -np.ones(len(ij_t), dtype='int32')
        for t, (ij, ik) in enumerate(zip(ij_t, ik_t)):
            for i in np.arange(first_neighbours[j_p[ij]],
                               first_neighbours[j_p[ij]+1]):
                if i_p[i] == j_p[ij] and j_p[i] == j_p[ik]:
                    jk_t[t] = i
                    break
        return ij_t, ik_t, jk_t
    else:
        return res


def find_indices_of_reversed_pairs(i_n, j_n, abs_dr_n):
    """Find neighbor list indices where reversed pairs are stored

    Given list of identifiers of neighbor atoms `i_n` and `j_n`,
    determines the list of indices `reverse` into the neighbor list
    where each pair is reversed, i.e. `i_n[reverse[n]]=j_n[n]` and
    `j_n[reverse[n]]=i_n[n]` for each index `n` in the neighbor list

    In the case of small periodic systems, one needs to be careful, because
    the same pair may appear more than one time, with different pair
    distances. Therefore, the pair distance must be taken into account.

    We assume that there is in fact one reversed pair for every pair.
    However, we do not check this assumption in order to avoid overhead.

    Parameters
    ----------
    i_n : array_like
       array of atom identifiers
    j_n : array_like
       array of atom identifiers
    abs_dr_n : array_like
        pair distances

    Returns
    -------
    reverse : numpy.ndarray
        array of indices into i_n and j_n
    """
    sorted_1 = np.lexsort(keys=(abs_dr_n, i_n, j_n))
    sorted_2 = np.lexsort(keys=(abs_dr_n, j_n, i_n))
    # np.testing.assert_equal(j_n[sorted_1], i_n[sorted_2])
    # np.testing.assert_equal(i_n[sorted_1], j_n[sorted_2])
    # np.testing.assert_equal(abs_dr_n[sorted_1], abs_dr_n[sorted_2])
    # print(np.c_[i_n[sorted_2], j_n[sorted_2], abs_dr_n[sorted_2],
    #            i_n[sorted_1], j_n[sorted_1], abs_dr_n[sorted_1]])
    tmp2 = np.arange(i_n.size)[sorted_2]
    tmp1 = np.arange(i_n.size)[sorted_1]
    # np.arange(i_n.size) are indices into the neighbor list, so
    #  * the nth element in tmp1 is the index where i,j was before reordering with sorted_1
    #  * the nth element in tmp2 is the index where j,i was before reordering with sorted_2
    reverse = np.empty(i_n.size, dtype=i_n.dtype)
    reverse[tmp1] = tmp2
    return reverse


def find_common_neighbours(i_n, j_n,  nat):
    """Find common neighbors of pairs of atoms

    For each pair ``(i1, j1)`` in the neighbor list, find all other pairs
    ``(i2, j1)`` which share the same ``j1``. This includes ``(i1,j1)``
    itself. In this way, create a list with ``n`` blocks of rows, where ``n``
    is the length of the neighbor list. All rows in a block have the same
    ``j1``. Each row corresponds to one triplet ``(i1, j1 ,i2)``. The number
    of rows in the block is equal to the total number of neighbors of ``j1``.

    Parameters
    ----------
    i_n : array_like
       array of atom identifiers
    j_n : array_like
       array of atom identifiers
    nat: int
        number of atoms

    Returns
    -------
    cnl_i1_i2: array
        atom numbers i1 and i2
    cnl_j1: array
        shared neighbor of i1 and i2
    nl_index_i1_j1: array
        index in the neighbor list of pair i1, j1
    nl_index_i2_j1: array
        index in the neighbor list of pair i2, j1

    Examples
    --------

    Accumulate random numbers for pairs with common neighbors:

    >>> import numpy as np
    >>> import matscipy
    >>> from ase.lattice.cubic import FaceCenteredCubic
    >>> from matscipy.neighbours import neighbour_list, find_common_neighbours
    >>> cutoff = 6.5
    >>> atoms = FaceCenteredCubic('Cu', size=[4, 4, 4])
    >>> nat = len(atoms.numbers)
    >>> print(nat)
    256
    >>> i_n, j_n, dr_nc, abs_dr_n = neighbour_list('ijDd', atoms, cutoff)
    >>> print(i_n.shape)
    (22016,)
    >>> cnl_i1_i2, cnl_j1, nl_index_i1_j1, nl_index_i2_j1 = find_common_neighbours(i_n, j_n, nat)
    >>> print(cnl_i1_i2.shape)
    (1893376, 2)
    >>> unique_pairs_i1_i2, bincount_bins = np.unique(cnl_i1_i2, axis=0, return_inverse=True) 
    >>> print(unique_pairs_i1_i2.shape)
    (65536, 2)
    >>> tmp = np.random.rand(cnl_i1_i2.shape[0])
    >>> my_sum = np.bincount(bincount_bins, weights=tmp, minlength=unique_pairs_i1_i2.shape[0]) 
    >>> print(my_sum.shape)
    (65536,)

    """
    # Create a copy of the neighbor list which is sorted by j_n, e.g.
    # +---------------+    +---------------+
    # | sorted by i_n |    | sorted by j_n |
    # +=======+=======+    +=======+=======+
    # | i_n   | j_n   |    | i_n   | j_n   |
    # +-------+-------+    +-------+-------+
    # | 1     | 2     |    | 2     | 1     |
    # +-------+-------+    +-------+-------+
    # | 1     | 95    |    | 4     | 1     |
    # +-------+-------+    +-------+-------+
    # | 2     | 51    |    | 81    | 2     |
    # +-------+-------+    +-------+-------+
    # | 2     | 99    |    | 12    | 2     |
    # +-------+-------+    +-------+-------+
    # | 2     | 1     |    | 6     | 2     |
    # +-------+-------+    +-------+-------+
    # | 3     | 78    |    | 143   | 3     |
    # +-------+-------+    +-------+-------+
    # | ...   | ...   |    | ...   | ...   |
    # +-------+-------+    +-------+-------+
    j_order = np.argsort(j_n)
    i_n_2 = i_n[j_order]
    j_n_2 = j_n[j_order]
    # Find indices in the copy where contiguous blocks with same j_n_2 start
    first_j = first_neighbours(nat, j_n_2) 
    num_rows_per_j = first_j[j_n+1] - first_j[j_n] 
    num_rows_cnl = np.sum(num_rows_per_j)

    # The common neighbor information could be stored as
    # a 2D array. However, multiple 1D arrays are likely
    # better for performance (fewer cache misses later).
    nl_index_i1_j1 = np.empty(num_rows_cnl, dtype=i_n.dtype)
    cnl_j1 = np.empty(num_rows_cnl, dtype=i_n.dtype)
    nl_index_i2_j1 = np.empty(num_rows_cnl, dtype=i_n.dtype)
    cnl_i1_i2 = np.empty((num_rows_cnl, 2), dtype=i_n.dtype)

    block_start = np.r_[0, np.cumsum(num_rows_per_j)]
    slice_for_j1 = {j1: slice(first_j[j1], first_j[j1+1]) for j1 in np.arange(nat)}
    for block_number, (i1, j1) in enumerate(zip(i_n, j_n)):
        slice1 = slice(block_start[block_number], block_start[block_number+1])
        slice2 = slice_for_j1[j1] 
        nl_index_i1_j1[slice1] = block_number
        cnl_j1[slice1] = j1 
        nl_index_i2_j1[slice1] = j_order[slice2]
        cnl_i1_i2[slice1, 0] = i1
        cnl_i1_i2[slice1, 1] = i_n_2[slice2]
    return cnl_i1_i2, cnl_j1, nl_index_i1_j1, nl_index_i2_j1


